//? cpmcompdef.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


// no double include guard needed since this occurs only as
// '#include <cpmcompdef.h>' in cpmdefinitions.h and the letter file
// is guarded against double inclusion.
/*************************************************************************
   cpmcompdef.h
   Description: In this file one defines how compiler options of the group
   -D influence the compilation directives defined in cpmdefinition.h
************************************************************************/
// ways to define R by compiler options of group -D
// one line in a makefile could for instance be
// defines = -D_CONSOLE -DCPM_RMP_C
// compilation than would result in an executable which uses a
// representation with 64 decimal places (actually the computed equivalent
// of binary places) for each number of type R.
// Since C+- is made compatible with C++20 it only supports a single way 
// to implement multiple precision arithmetics: based on 
// boost::multiprecision . Unfortunately this is not fully in harmony with
// using Eigen library. It only works if compilation is done under
// the -fpermissive option. If one likes not be overoaded with warnings one
// also needs the -w option. Also one can't use expression templates in Eigen
// (the flag which regulates this is set in the C+- code).

#if defined(CPM_RFLOAT)  // R float
   #undef CPM_FLOAT
   #undef CPM_DOUBLE 
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP   
   #define CPM_FLOAT
      
#elif defined(CPM_R)  // R double
   #undef CPM_FLOAT
   #undef CPM_DOUBLE 
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_DOUBLE
   
#elif defined(CPM_RDOUBLE)  // R double (same as above, for uniformity)
   #undef CPM_FLOAT
   #undef CPM_DOUBLE 
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_DOUBLE
   
#elif defined(CPM_RLONG)  // R long double
   #undef CPM_FLOAT
   #undef CPM_DOUBLE 
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP   
   #define CPM_LONG      
   
#elif defined(CPM_RQUAD)  // R quad
   #undef CPM_FLOAT
   #undef CPM_DOUBLE 
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP   
   #define CPM_QUAD  
         
#elif defined(CPM_RMP) // R multiprecision, here comparable with double
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 16
   
#elif defined(CPM_RMP_A)
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 24
   
#elif defined(CPM_RMP_B) // R multiprecision, here comparable with quad
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 32
   
#elif defined(CPM_RMP_C)
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 64
   
#elif defined(CPM_RMP_D)
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 128
   
#elif defined(CPM_RMP_E)
   #undef CPM_FLOAT
   #undef CPM_DOUBLE
   #undef CPM_LONG
   #undef CPM_QUAD
   #undef CPM_MP
   #define CPM_MP 256
   
#endif

