//? cpmdiscretespace.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_DISSPC_H_
#define CPM_DISSPC_H_
/*
    Description: discrete space for defining fields
      that can be added together in arbitrary number.
      Note that adding my standard fields accumulates data
      that may create stack overflow for too many terms
      (~3000 in one case)
*/
#include <cpmdigraph.h>
#include <cpmf.h>

namespace CpmGeo{

   using CpmArrays::V;
   using CpmArrays::Vr;
   using CpmFunctions::F;
   #ifdef CPM_Fn
   using CpmFunctions::F1;
   #endif

   template<class X>
class DisSpc: public DiGraph<X>{
      // directed graph with position descriptors attached to vertices
      // Has nothing to do with CpmGraphics::Graph

      typedef DisSpc Type;
      typedef DiGraph<X> Base;
   public:
      virtual DiGraph<X>* clone()const{ return new DisSpc(*this);}

explicit DisSpc(const V<X>& vert=V<X>()):
      DiGraph<X>(vert){}
         // a graph with n vertices and no edges
   };

// discrete fields
// arithmetic is defined on a common grid.
// What is the right place for the grid?
// Possibilities:
// 1. template argument of Type DisSpc
// 2. derived from DisSpc
// 3. static member of type DisSpc ; this is chosen here!

template <class X, class Y>
class DisField{ // discrete field
   typedef DisField Type;
   typedef Y ScalarType;
   static DisSpc<X> grid;
   Vr<Y> y;

public:
   CPM_IO_V
   static void setGrid(const V<X> vx){ grid=DisSpc<X>(vx);}
   static Z card(){ return grid.card();}
   void operator+=(const DisField& f){ y+=f.y;}
   DisField(){y=V<Y>(grid.car(),Y());}
   explicit DisField(const F<X,Y>& f);
   F<X,Y> contField()const;
   void setY(const V<Y>& yIn){ y=yIn;}
};

template <class X, class Y>
DisSpc<X> DisField<X,Y>::grid;


template<class X,class Y>
DisField<X,Y>::DisField(const F<X,Y>& f)
{
   y=CpmFunctions::apply<X,Y>(f,grid.get_x());
}

template<class X,class Y>
F<X,Y> DisField<X,Y>::contField()const
{
 //  Z n=grid.car();
   return F<X,Y>(new CpmAlgorithms::
      AveragingInterpolator<X,Y>(grid.get_x(),y));
}

template<class X,class Y>
bool DisField<X,Y>::prnOn(ostream& str)const
{
   cpmp(y);
   return true;
}

template<class X,class Y>
bool DisField<X,Y>::scanFrom(istream& str)
{
   cpms(y);
   return true;
}

} // namespace

#endif
