//? cpmfontgeneration.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmfontgeneration.h>

using CpmFonts::PrlLet;
using CpmFonts::Let;
using CpmFonts::PrlFont;
using CpmFonts::Font;

// other helpers:

using CpmRoot::Z;
using CpmRoot::Word;
using CpmRoot::L;
using CpmRoot::R;

using CpmSystem::IFileStream;
using CpmSystem::OFileStream;

using CpmLinAlg::R2;

using CpmArrays::IvZ;
using CpmArrays::V;
using CpmArrays::M;
using CpmArrays::LEAN;

using CpmFunctions::F;
#ifdef CPM_Fn
using CpmFunctions::F1;
#endif

using CpmGraphics::Viewport;
using CpmGraphics::Img24;
using CpmGraphics::ColRef;
using CpmGraphics::rgb;
using CpmGraphics::xy;
using namespace std;
using namespace std::placeholders;

namespace{
bool printFunc(ColRef const& cr, ColRef const& sep)
{
   if (cr.r>sep.r && cr.g>sep.g && cr.b>sep.b) return false;
   else return true;
}
} // namespace

////////////////////// class PrlLet ////////////////////////////////////

PrlLet::PrlLet(Img24 const& imr, ColRef sep)
{
   Z mL=3;
   static Word loc("PrlLet(Image<R>)");
   CPM_MA
   w_=imr.w();
   h_=imr.h();
   r_=V< V<IvZ> >(h_, V<IvZ>(),0);
// We allow for the case that the image was taken with an
// electronic camera (with camera aligned properly) but
// sufficient resolution that allows proper bi-level
// binarization of the values.
#ifdef CPM_Fn
   F<ColRef,bool> f=F1<ColRef,ColRef,bool>(sep)(printFunc);
#else
   F<ColRef,bool> f(bind(printFunc,_1,sep));
#endif
   for (Z i=0;i<h_;++i){
       r_[i]=imr[i].valOn(f);
         // this is the new member functon V<>::valOn(...)
       r_[i].b_(0);
   }
   CPM_MZ
}

namespace{
bool printF(ColRef const& cr, R const& sep)
{
   if (sep < cr.getR()) return false; // only very dark parts
      // will be printed
   if (sep < cr.getG()) return false;
   if (sep < cr.getB()) return false;
   return true;
}
} // namespace

PrlLet::PrlLet(Word const& fn)
{
   Z mL=1;
   Word loc("PrlLet("&fn&")");
   CPM_MA
   Img24 img(fn);
  // if (cpmverbose>mL) img.vis();
   cpmwait(4);
   w_=img.w();
   h_=img.h();
   r_=V< V<IvZ> >(h_, V<IvZ>(),0);
// We allow for the case that the image was taken with an
// electronic camera (with camera aligned properly) but
// sufficient resolution that allows proper bi-level
// binarization of the values.
   R sep=127;
#ifdef CPM_Fn
   F<ColRef,bool> f=F1<ColRef,R,bool>(sep)(printF);
#else
   F<ColRef,bool> f(std::bind(printF,_1,sep));
#endif
   for (Z i=0;i<h_;++i){
      r_[i]=img[i].valOn(f);
         // this is the new member functon V<>::valOn(...)
      r_[i].b_(0);
   }
   CPM_MZ
}

namespace{
bool dimFunc(Z const& i){ return i>0;}
}

V<Z> minMax(V<Z> const& v)
{
   Z n=v.dim();
   if (n==0) return V<Z>(0);
   Z vMin=v.fir();
   Z vMax=vMin;
   for (Z i=v.b()+1;i<=v.e();++i){
      Z vi=v[i];
      if (vi<vMin) vMin=vi;
      else if (vi>vMax) vMax=vi;
      else ;
   }
   return V<Z>{vMin,vMax};
}

PrlLet PrlLet::min()const
{
   Z mL=2;
   static Word loc("PrlLet::min()");
   CPM_MA
   if (h_==0){
      cpmmessage("trivial exit for h_==0");
      CPM_MZ
      return *this;
   }
   V<Z> dims(h_,LEAN);
   Z i;
   for (i=0;i<h_;++i) dims[i]=r_[i].dim();
   V<IvZ> roi=dims.valOn(F<Z,bool>(dimFunc));
   IvZ iv1=roi.fir();
   Z iFirst=iv1.fir();
   IvZ iv2=roi.last();
   Z iLast=iv2.last();
   Z h=1+iLast-iFirst;
   V< V<IvZ> > r(h,LEAN);
   Z iRead=iFirst;
   for (i=0;i<h;++i) r[i]=r_[iRead++];
   V<Z> firsts(0,LEAN);
   V<Z> lasts(0,LEAN);
   for (i=0;i<h;++i){
      V<IvZ> ri=r[i];
      Z di=ri.dim();
      if (di>0){
         firsts&=ri.fir().fir();
         lasts&=ri.last().last();
      }
   }
   V<Z> mmf=minMax(firsts); // min is needed
   V<Z> mml=minMax(lasts); // max is needed
   Z f=mmf.fir();
   Z l=mml.last();
   Z j,w=1+l-f;
   for (i=0;i<h;++i){
      Z di=r[i].dim();
      for (j=0;j<di;++j) r[i][j]-=f;
   }
   CPM_MZ
   return PrlLet(r,h,w);
}

Img24 PrlLet::toImg24()const
{
   Z mL=1;
   Word loc("PrlLet::toImg24()");
   CPM_MA
   Z c1=255,c0=0;
   rgb cw(c1,c1,c1);
   rgb cb(c0,c0,c0);
   Img24 res(w_,h_,cw);
   Z i,j;
   for (i=0;i<h_;++i){
      V<IvZ> ri=r_[i];
      Z di=ri.dim();
      for (j=0;j<di;++j){
         IvZ vij=ri[j];
         if (!vij.isVoid()){
            Z x1=vij.fir();
            Z x2=vij.last();
            Z y1=i,y2=y1;
            res.line(x1,y1,x2,y2,cb);
         }
      }
   }
   CPM_MZ
   return res;
}

Img24 PrlLet::toImg24(Z hl, Z hu, Z r)const
{
   Z mL=3;
   Word loc("PrlLet::toImg24(Z,Z,Z)");
   CPM_MA
   R c1=255, c0=0;
   rgb cw(c1,c1,c1);
   rgb cb(c0,c0,c0);
   Z wf=w_+2*r;
   Z hf=hl+hu;
   Img24 im=toImg24();
   Img24 res(wf,hf,cw);
   xy p(r, hu-h_+d_);
   res.insert(p,im);
   CPM_MZ
   return res;
}

void PrlLet::print(Z x, Z y, Img24& im)const
{
   Z i,j;
   R c0=0,c1=100;
   rgb cb(c0,c1,c0);
   Z xShift=x;
   Z yShift=y-h_+d_;
   for (i=0;i<h_;++i){
      V<IvZ> ri=r_[i];
      Z di=ri.dim();
      for (j=0;j<di;++j){
         IvZ vij=ri[j];
         Z x1=vij.fir()+xShift;
         Z x2=vij.last()+xShift;
         Z y1=i+yShift,y2=y1;
         im.line(x1,y1,x2,y2,cb);
      }
   }
}

bool PrlLet::prnOn(ostream& str)const
{
   cpmwt("PrlLet");
   cpmp(r_);
   cpmp(h_);
   cpmp(w_);
   cpmp(d_);
   cpmp(c_);
   return true;
}

bool PrlLet::scanFrom(istream& str)
{
   cpms(r_);
   cpms(h_);
   cpms(w_);
   cpms(d_);
   cpms(c_);
   return true;
}

// The following commentarized version was useful during development.
/*******************
bool PrlLet::scanFrom(istream& str)
{
   bool b;
   b=CpmRoot::scanT(r_,str);
   if (!b){
      cpmcerr<<"Problem with reading r_"<<endl;
      return b;
   }
   b=CpmRoot::scanT(h_,str);
   if (!b){
      cpmcerr<<"Problem with reading h_"<<endl;
      return b;
   }
   b=CpmRoot::scanT(w_,str);
   if (!b){
      cpmcerr<<"Problem with reading w_"<<endl;
      return b;
   }
   b=CpmRoot::scanT(d_,str);
   if (!b){
      cpmcerr<<"Problem with reading d_"<<endl;
      return b;
   }
   b=CpmRoot::scanT(c_,str);
   if (!b){
      cpmcerr<<"Problem with reading c_"<<endl;
      return b;
   }
   return true;
}
***********************/

//////////////////////// class Let /////////////////////////////////////

Let::Let(PrlLet const& pl, Img24 const& im, R frac)
{
   Z mL=3;
   Word loc("Let(PrlLet,Img24,R)");
   CPM_MA
   c_=pl.c();
   Z hi=im.h();
   cpmassert(hi>0,loc);
   Z wi=im.w();
   hi=cpmrnd(hi*frac);
   wi=cpmrnd(wi*frac);
   Z bilinear=1;
   r_=im.interpolate(wi,hi,bilinear);
   CPM_MZ
}

void Let::print(Z x, Z y, Img24& im , ColRef cr)const
{
   xy cori=im.cor();
   xy p(x,y);
   im.insert(p,r_,cr);
   xy corf=im.cor();
   cpmassert(cori.x==corf.x,"Let::print(...)");
   cpmassert(cori.y==corf.y,"Let::print(...)");
}

bool Let::prnOn(ostream& str)const
{
   cpmwt("Let begin");
   cpmp(r_);
   cpmp(c_);
   cpmwt("Let end");
   return true;
}

bool Let::scanFrom(istream& str)
{
   cpms(r_);
   cpms(c_);
   return true;
}

//////////////////////// class PrlFont /////////////////////////////////

V<PrlLet> PrlFont::fromPPM(Word const& dirIn, Word const& fnOut)
{
   Z mL=2;
   Word loc("PrlFont::fromPPM(Word,Word)");
   CPM_MA

   Z df=91;
   V<Word> vw(df,LEAN);
   vw[0]="0";
   vw[1]="1";
   vw[2]="2";
   vw[3]="3";
   vw[4]="4";
   vw[5]="5";
   vw[6]="6";
   vw[7]="7";
   vw[8]="8";
   vw[9]="9";

   vw[10]="a";
   vw[11]="b";
   vw[12]="c";
   vw[13]="d";
   vw[14]="e";
   vw[15]="f";
   vw[16]="g";
   vw[17]="h";
   vw[18]="i";
   vw[19]="j";
   vw[20]="k";
   vw[21]="l";
   vw[22]="m";
   vw[23]="n";
   vw[24]="o";
   vw[25]="p";
   vw[26]="q";
   vw[27]="r";
   vw[28]="s";
   vw[29]="t";
   vw[30]="u";
   vw[31]="v";
   vw[32]="w";
   vw[33]="x";
   vw[34]="y";
   vw[35]="z";

   Z i,s0=26;
   for (i=36;i<62;++i) vw[i]=vw[i-s0]+"2";
      // the corresponding capital letters get a '2'
      // appended

   vw[62]="20";
   vw[63]="21";
   vw[64]="23";
   vw[65]="24";
   vw[66]="25";
   vw[67]="26";
   vw[68]="28";
   vw[69]="29";
   vw[70]="2a";
   vw[71]="2b";
   vw[72]="2c";
   vw[73]="2d";
   vw[74]="2e";
   vw[75]="2f";
   vw[76]="3a";
   vw[77]="3b";
   vw[78]="3c";
   vw[79]="3d";
   vw[80]="3e";
   vw[81]="3f";
   vw[82]="40";
   vw[83]="5b";
   vw[84]="5c";
   vw[85]="5d";
   vw[86]="7b";
   vw[87]="7c";
   vw[88]="7d";
   vw[89]="7e";
   vw[90]="2ds";

   V<L> c(df,LEAN);
   for (i=0;i<36;++i) c[i]=vw[i][1]; // first character of word
   c[36]='A';
   c[37]='B';
   c[38]='C';
   c[39]='D';
   c[40]='E';
   c[41]='F';
   c[42]='G';
   c[43]='H';
   c[44]='I';
   c[45]='J';
   c[46]='K';
   c[47]='L';
   c[48]='M';
   c[49]='N';
   c[50]='O';
   c[51]='P';
   c[52]='Q';
   c[53]='R';
   c[54]='S';
   c[55]='T';
   c[56]='U';
   c[57]='V';
   c[58]='W';
   c[59]='X';
   c[60]='Y';
   c[61]='Z';

   c[62]='\x20';
   c[63]='\x21';
   c[64]='\x23';
   c[65]='\x24';
   c[66]='\x25';
   c[67]='\x26';
   c[68]='\x28';
   c[69]='\x29';
   c[70]='\x2a';
   c[71]='\x2b';
   c[72]='\x2c';
   c[73]='\x2d';
   c[74]='\x2e';
   c[75]='\x2f';
   c[76]='\x3a';
   c[77]='\x3b';
   c[78]='\x3c';
   c[79]='\x3d';
   c[80]='\x3e';
   c[81]='\x3f';
   c[82]='\x40';
   c[83]='\x5b';
   c[84]='\x5c';
   c[85]='\x5d';
   c[86]='\x7b';
   c[87]='\x7c';
   c[88]='\x7d';
   c[89]='\x7e';
   c[90]='_';
   Z n=c.dim();
   V<PrlLet> res(n,LEAN);
    for (i=0;i<n;++i){
      L ci=c[i];
      PrlLet pi; // trivial
      if (ci != '\x20'){ // not the blank character
         // there is no ppm-image of a blank field
         Word fni(dirIn&"f"&vw[i]&".ppm");
         pi=PrlLet(fni);
         pi=pi.min();
         if (cpmverbose>mL){
            Img24 imi=pi.toImg24();
            imi.vis();
            cpmwait(1);
         }
      }
      pi.setMen(c[i]);
      res[i]=pi;
   }
   if (fnOut.dim()>0){
      OFileStream out(fnOut);
      res.prnOn(out());
   }
   CPM_MZ
   return res;
}

PrlFont::PrlFont( V<PrlLet> const& l):
   l_(l),im_(l.dim(),LEAN),h_(0),hu_(0),hl_(0)
{
   Z mL=1;
   Word loc("PrlFont(Vl<PrlLet>)");
   CPM_MA
   makeReady();
      // now all size descriptors of the PrlLets have been set
   Z d=l_.dim(),hi,di,hui,hli;
   for (Z i=0;i<d;++i){
      hi=l_[i].h();
      di=l_[i].d();
      hli=di;
      hui=hi-di;
      if (hi>h_) h_=hi;
      if (hli>hl_) hl_=hli;
      if (hui>hu_) hu_=hui;
   }
   R r=0.5+0.05*h_;
   R c1=255;
   for (Z i=0;i<d;++i){
      cpmmessage("starting i="&cpm(i));
      L ci=l_[i].c();
      if (ci=='\x20')
         im_[i]=Img24(l_[i].w(),hl_+hu_,rgb(c1,c1,c1));
      else
         im_[i]=l_[i].toImg24(hl_,hu_,cpmtoz(r));
   }
   CPM_MZ
}

PrlFont::PrlFont(Word const& fileName)
{
   IFileStream in(fileName);
   scanFrom(in());
}

bool PrlFont::prnOn(ostream& str)const
{
   cpmwt("PrlFont");
   cpmp(l_);
   cpmp(im_);
   cpmp(h_);
   cpmp(hl_);
   cpmp(hu_);
   return true;
}


bool PrlFont::scanFrom(istream& str)
{
   cpms(l_);
   cpms(im_);
   cpms(h_);
   cpms(hl_);
   cpms(hu_);
   return true;
}

void PrlFont::makeReady()
      // setting all descenders consistently
{
   Z mL=1;
   static Word loc("PrlFont::makeReady()");
   CPM_MA
   Z j=getIntNum('j');
   Z g=getIntNum('g');
   Z p=getIntNum('p');
   Z q=getIntNum('q');
   Z y=getIntNum('y');
   Z quc=getIntNum('Q');

   Z comma=getIntNum('\x2c');
   Z blank=getIntNum('\x20');
   Z minus=getIntNum('\x2d');
   Z mult=getIntNum('\x2a');
   Z semicol=getIntNum('\x3b');

   l_[j].setDes(90); // more like y
   l_[g].setDes(82);
   l_[p].setDes(80);
   l_[q].setDes(80);
   l_[y].setDes(92);
   l_[quc].setDes(74);
   l_[comma].setDes(82); // surprising but true
   l_[minus].setDes(-100);
   l_[mult].setDes(-40);
   l_[semicol].setDes(82);
   l_[blank].setWid(110); // set width of blank
   l_[getIntNum('\x24')].setDes(87); // $ otherwise much too high
   l_[getIntNum('\x28')].setDes(60); // brackets
   l_[getIntNum('\x29')].setDes(60);
   l_[getIntNum('\x5b')].setDes(60);
   l_[getIntNum('\x5d')].setDes(60);
   l_[getIntNum('\x7b')].setDes(60);
   l_[getIntNum('\x7d')].setDes(60);
   l_[getIntNum('\x7e')].setDes(-100); // tilde ~
   l_[getIntNum('\x3d')].setDes(-80); // =
//   Z d=l_.dim();
//   for (Z i=0;i<d;++i){
//     Z wi=l_[i].w();
//   }
   CPM_MZ
}

Z PrlFont::getIntNum(L c)const
{
   static bool firstrun=true;
   static M<L,Z> st; // search table
   Z d=l_.dim();
   if (firstrun){
      for (Z i=0;i<d;++i){
         PrlLet pi=l_[i];
         L ci=pi.c();
         st[ci]=i;
      }
      firstrun =false;
   }
   return st[c];
}

////////////////////////// class Font //////////////////////////////////

bool Font::prnOn(ostream& str)const
{
   Z mL=3;
   static Word loc("Font::prnOn(ostream)");
   CPM_MA
   cpmwt("Font");
   cpmp(l_);
   cpmp(h_);
   cpmp(w_);
   cpmp(mw_);
   cpmp(st_);
   CPM_MZ
   return true;
}

bool Font::scanFrom(istream& str)
{
   Z mL=3;
   static Word loc("Font::scanFrom(istream)");
   CPM_MA
   cpms(l_);
   cpms(h_);
   cpms(w_);
   cpms(mw_);
   cpms(st_);
   CPM_MZ
   return true;
}

Font::Font( PrlFont const& pf, Z hAim)
{
   Z mL=2;
   static Word loc("Font(PrlFont,Z)");
   cpmstatus("making font "&cpm(hAim),2);
   CPM_MA
   Z i,pd=pf.dim();
   if (pd<1) return;
   l_=V<Let>(pd,LEAN);
   Z ph=pf.h();
   cpmassert(ph>0,loc);
   R hInv=R(1.)/ph;
   R frac=hAim*hInv;

   Z wMin{0},wMax{0};
   R mw=0;
   for (i=0;i<pd;++i){
      cpmmessage("loop: i="&cpm(i));
      PrlLet pli=pf.getLet(i);
      l_[i]=Let(pli,pf.getImg(i),frac);
      L ci=l_[i].c();
      st_[ci]=i;
      Z wi=l_[i].w();
      if (i==0){
         h_=l_[i].h(); // all li.h() are equal by construction
         wMin=wi;
         wMax=wi;
         mw=wi;
      }
      else{
         mw+=wi;
         if (wi>wMax) wMax=wi;
         if (wi<wMin) wMin=wi;
      }
   }
   w_=IvZ(wMin,wMax);
   mw_=mw/pd;
   cpmstatus("font "&cpm(hAim)&" ready ",2);
   CPM_MZ
}

Font::Font(Word const& fn)
{
   Z mL=3;
   static Word loc("Font(Word)");
   CPM_MA
   IFileStream in(fn);
   if (!in()) cpmerror("opening file "&fn&" failed");
   Font fTemp;
   bool suc=fTemp.scanFrom(in());
   if (!suc) cpmerror("could not read Font from "&fn);
   *this=fTemp;
   CPM_MZ
}

void Font::show(Img24& im , ColRef cr)const // write a font on im
{
   Z mL=2;
   static Word loc("Font::show(Img24&)");
   CPM_MA
   IvZ wf=w();
   Z hf=h();
   Z xStart=wf.inf();
   Z yStart=hf;
   Z pd=dim();
   Z sx=xStart;
   Z sy=yStart;
   Z yShift=hf+hf/2;
   xy cori=im.cor();
   Z wi=im.w()-wf.sup();
   for (Z i=0;i<pd;++i){
      Let li=l_[i];
      li.print(sx,sy,im,cr);
      sx+=li.w();
      if (sx>=wi){
         sx=xStart;
         sy+=yShift;
      }
      im+=cori;
      im.vis();
         // if the arguments were missing one would not
         // see the result in the same window as the original im
   }
   CPM_MZ
}

void Font::wrt(Word const& w, xy& p, Img24& img, ColRef cr)const
{
   Z x=p.x, y=p.y;
   for (Z i=w.b();i<=w.e();++i){
      L ci=w[i];
      Let li=getLet(ci);
      li.print(x,y,img,cr);
      x+=li.w();
   }
   p=xy(x,y);
}

Z Font::len(Word const& w)const
{
   Z res=0;
   for (Z i=w.b();i<=w.e();++i){
      L ci=w[i];
      Let li=getLet(ci);
      res+=li.w();
   }
   return res;
}

Font Font::factory(Z fHeight)
{
   Z mL=2;
   Word loc("Font::factory(Z)");
   CPM_MA
   Word fLoc="fonts/fontshapes1.txt"; // relative path OK
   IFileStream in(fLoc);
   V<PrlLet> vpl(0,LEAN);
   vpl.scanFrom(in());
   PrlFont pft(vpl);
   CPM_MZ
   return Font(pft,fHeight);
}

V<Font> Font::massFactory(Word const& fnIn, Word const& fnOut)
{
   Z mL=1;
   Word loc="Font::massFactory(";
   loc&=(fnIn&","&fnOut&")");
   CPM_MA
   IFileStream in(fnIn);
   V<PrlLet> vpl;
   vpl.scanFrom(in());
   PrlFont pft(vpl);
   Z n=6;
   V<Font> res(n,LEAN);
   V<Z> heights(n,LEAN);
   heights[0]=10;
   heights[1]=12;
   heights[2]=14;
   heights[3]=16;
   heights[4]=18;
   heights[5]=24;
   for (Z i=0;i<n;++i) res[i]=Font(pft,heights[i]);
   OFileStream ofs(fnOut);
   bool suc=res.prnOn(ofs());
   if (suc) cpmmessage(mL, loc&": writing "&fnOut&" done");
   CPM_MZ
   return res;
}

Z Font::com(CpmFonts::Font const& ft) const
{ return CpmRoot::Root<Z>(h_).com(ft.h_);}
