//? cpmimagingtools.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmangle.h>
#include <cpmimagingtools.h>
#include <cpmzinterval.h>
#include <cpmrfunc.h>
#include <cpmsystemdependencies.h>

using namespace CpmRoot;
using namespace CpmGeo;
using namespace CpmSystem;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmImaging;

///////////////////////////// Color //////////////////////////////////
//CPM_STEREOLEFT_R
#if defined(CPM_STEREOLEFT_R)
   Z CpmImaging::stereoLeftR = CPM_STEREOLEFT_R;
#else
   Z CpmImaging::stereoLeftR = 255;
#endif

#if defined(CPM_STEREOLEFT_G)
   Z CpmImaging::stereoLeftG = CPM_STEREOLEFT_G;
#else
   Z CpmImaging::stereoLeftG = 0;
#endif

#if defined(CPM_STEREOLEFT_B)
   Z CpmImaging::stereoLeftB = CPM_STEREOLEFT_B;
#else
   Z CpmImaging::stereoLeftB = 0;
#endif

#if defined(CPM_STEREORIGHT_R)
   Z CpmImaging::stereoRightR = CPM_STEREORIGHT_R;
#else
   Z CpmImaging::stereoRightR = 0;
#endif

#if defined(CPM_STEREORIGHT_G)
   Z CpmImaging::stereoRightG = CPM_STEREORIGHT_G;
#else
   Z CpmImaging::stereoRightG = 255;
#endif

#if defined(CPM_STEREORIGHT_B)
   Z CpmImaging::stereoRightB = CPM_STEREORIGHT_B;
#else
   Z CpmImaging::stereoRightB = 255;
#endif

R Color::gamma=1.;
// should be 1

R Color::valMax_=256-1e-6;

bool Color::prnOn(ostream& str)const
{
   using CpmRoot::write;
   CpmRoot::writeTitle("Color",str);
   cpmp(r);
   cpmp(g);
   cpmp(b);
   return true;
}

bool Color::scanFrom(istream& str)
{
   cpms(r);
   cpms(g);
   cpms(b);
   return true;
}

Z Color::com(const Color& c)const
{
   if (r<c.r) return 1;
   if (r>c.r) return -1;
   if (g<c.g) return 1;
   if (g>c.g) return -1;
   if (b<c.b) return 1;
   if (b>c.b) return -1;
   return 0;
}

Color Color::operator+(const Color& c)const
{
   return Color(r+c.r,g+c.g,b+c.b);
}

Color Color::operator*(const R& x)const
{
   return Color(R(r)*x,R(g)*x,R(b)*x);
}

void Color::make(Palette c)
   {
// The Palette colors make use only of a few code values
      static const Z b4=255;
      static const Z b34=215;
      static const Z b3=191;
      static const Z b2=127;
      static const Z b1=66;
      static const Z b0=0;
      static const Z bright=200;
      static const Z dark=50;
      switch(c){
// dark colors
      case BLACK:       r=b0; g=b0; b=b0; break; // experiment
      case BLUE:        r=b0; g=b0; b=b4; break;
      case GREEN:       r=b0; g=b4; b=b0; break;
      case CYAN:        r=b0; g=b2; b=b2; break;
      case RED:         r=b4; g=b0; b=b0; break;
      case MAGENTA:     r=b2; g=b0; b=b2; break;
      case BROWN:       r=b2; g=b2; b=b0; break;
      case FAINTGRAY:   r=b3; g=b3; b=b3; break;
         // closer to white than LIGHTGRAY
      case FAINTYELLOW: r=b34; g=b34; b=b3; break;
         // for friendly background only

      case FAINTGREEN: r=b34; g=b4; b=b34; break;
         // for friendly background only

      case LIGHTGRAY:   r=b2; g=b2; b=b2; break;
         // = 'dark white'
// darker colors
      case DARKBLUE:    r=b0; g=b0; b=b2; break;
      case DARKGREEN:   r=b0; g=b2; b=b0; break;
      case DARKCYAN:    r=b0; g=b1; b=b1; break;
      case DARKRED:     r=b2; g=b0; b=b0; break;
      case DARKMAGENTA: r=b1; g=b0; b=b1; break;
      case DARKGRAY:    r=b1; g=b1; b=b1; break; // = 'light black'
// bright colors
      case LIGHTBLUE:   r=b2; g=b2; b=b4; break;
      case LIGHTGREEN:  r=b2; g=b4; b=b2; break;
      case LIGHTCYAN:   r=b2; g=b4; b=b4; break;
      case LIGHTRED:    r=b4; g=b2; b=b2; break;
      case LIGHTMAGENTA:r=b4; g=b2; b=b4; break;
      case YELLOW:      r=b4; g=b4; b=b2; break; // light brown
      case ORANGE:      r=b4; g=b2; b=b0; break;
      case WHITE:       r=b4; g=b4; b=b4; break;
// specific colors for anaglyphic stereo representation
  //    case STEREOBLACK: r=87; g=31; b=b0; break; // black values
  // according to
         // experiments of Thomas Dera 2000-9-13
      case STEREOBLACK: r=0; g=0; b=0; break; // dark values
      // but not necessarily really black. r=0 is necessary
      // however
      case STEREOLEFT: r=stereoLeftR; g=stereoLeftG; b=stereoLeftB; break;

      case STEREORIGHT: r=stereoRightR; g=stereoRightG;
                           b=stereoRightB; break;
         // Originally I had STEREOLEFT =(255,0,0),
         // STEREORIGHT=(0,255,255).
         // Then:
         // Proposal by Bernhard Mutze (2009-06-19) to have no green,
         // only blue for better right-left separation with my Apramax
         // filter 'glasses'. Then red must be a bit reduced (b3 seems OK)
         // This gave the impetus to make things general.
// not a color
      case INVISIBLE:   r=-1; g=-1; b=-1; break;
      case STEREO:      r=-1; g=-1; b=-1; break; // same as INVISIBLE
         // even to me my old name INVISIBLE suggested the wrong
         // association after having not worked with it for a while.
         // could also be named ANAGLYPH
      case TRANSPARENT_: r=-137; g=-1; b=-1; break;
      case WRITECOLOR:   r=dark; g=dark; b=dark; break;
      case GRIDCOLOR:    r=bright; g=bright; b=bright; break;
         // should not appear in the previous list, so that curves
         // normally don't get the same color as the grid
      default:  r=0; g=0; b=0; break;
      }
   }

   namespace{
      const Z nUsed=17;
      // we use the first nUsed colors of Palette
   }

   Color::Color(Z colNumber)
   {
      Z mL=4;
      static Word loc("Color(Z)");
      CPM_MA
      Z i=colNumber-1;
      if (i<0) { r=-1; g=-1; b=-1;}
      else{ // else was missing till 2005-04-06
         i=i%nUsed;
            // now i in {0,1,....nUsed-1}
         make((Palette)(i));
      }
      CPM_MZ
   }

   Z Color::toZ(Palette p)
   {
      Z res=0;
      if (p!=INVISIBLE && p!=STEREO){
         res=1+Z(p);
         if (res>nUsed) cpmwarning(
            "Color::toZ(Palette): colors beyond WHITE not expected here")\
            ;
      }
      return res;
   }

   Palette Color::toPal(Z i)
   {
      Z j=i-1;
      if (j<0) return STEREO;
      else return Palette(j%nUsed);
   }

namespace{ // anonymous
   const R xg=-0.5;
   const R yg=0.5*cpmsqrt(3.);
   const R2 er(1,0);
   const R2 eg(xg,yg);
   const R2 eb(xg,-yg);
      // er+eg+eb=0,
      // er.abs()==eg.abs()==eb.abs()==1
}

Color::Color(C const& z, Iv mm)
// no trigonometric functions called, fast
   {
      R2 v(z.real(),z.imag());
      R vr=v.normalize_();
      R r_=((v|er)+1.)*0.5;
      R g_=((v|eg)+1.)*0.5;
      R b_=((v|eb)+1.)*0.5;
      // r_+g_+b_=3/2, 0<=r_,g_,b_<=1
      R rho=ImagingTools::response(
         mm.inf(),mm.sup(),0,255, gamma,vr);
         // gamma is Color::gamma
      r=cpmtoz(rho*r_);
      g=cpmtoz(rho*g_);
      b=cpmtoz(rho*b_);
   }

   Color Color::hue(R const& h, bool finalMagenta)
   {
      static R fac=R(7)/R(8);
      R phi=cpm2pi*h;
      if (finalMagenta) phi*=fac;
      C z=C::expi(phi);
      R gammaMem=Color::gamma;
      Color::gamma=0.;
      Color res(z);
      Color::gamma=gammaMem;
      return res/*.brighten()*/;
   }

   Color Color::brighten(R const& b)const
   {
      Z m=max(), up=ColRef::upperLimit;
      if (m==0) return Color(up,up,up);
      R fac=b*R(up)/R(m);
      return Color(fac*r,fac*g,fac*b);
   }

   Color::Color(R r, Iv mm)
   {
      static Color cL(RED);
      static Color c0(WHITE);
      static Color cU(BLUE);
      R p=(mm.ramp(r)-0.5)*2;
      Color res;
      if (p<0){
         p*=-1;
         res=p*cL+(R(1)-p)*c0;
      }
      else{
         res=p*cU+(R(1)-p)*c0;
      }
      *this=res;
   }

   Color::Color(R3 const& y)
   {
      static const Iv iv(0,1);
      static const Z n=256;
      static const bool restrict=true; // named quantities
         // support understanding. Using literals as function
         // arguments falls short in this respect.
      r=iv.qun(y[1],n,restrict)-1;
      g=iv.qun(y[2],n,restrict)-1;
      b=iv.qun(y[3],n,restrict)-1;
   }

   R3 Color::toR3()const
   {
      static const IvZ ivz(Z(0),Z(255));
      static const Iv iv(0,1);
      return R3(iv.qunInv(r,ivz),iv.qunInv(g,ivz),iv.qunInv(b,ivz));
   }

///////////////////////////// ImagingTools ///////////////////////////

Z ImagingTools::verbose=0;

R ImagingTools::fMono(R_Vector const& a)
{ return a.mean();}

R ImagingTools::fMax(R_Vector const& a)
{ return a.sup();}

R ImagingTools::fMin(R_Vector const& a)
{ return a.inf();}

R ImagingTools::int2(R f11, R f12, R f21, R f22, R p, R q, Z method)
// int2 stands for interpolation in 2 dimensions.
// The returned value is a weighted meen of the fij. Notice
// p+p_=1, q+q_=1, which leads to p*q_+p_*q_+p*q+p_*q=1.
// For interpretation: p and q follow the x and y axis respectively.
// The corners of the rectangle in which the interpolation point is
// placed and to which the reduced coordinates p and q refer are as
// follows:
//
//      f11      f12
//
//      f21      f22
//
// Method=1 means bilinear interpolation and method=0 means
// pixelation. The latter is useful if the number of pixels is to be
// compared to expectations.
{
   if (method==1){ // bilinear interpolation
      R p_=R(1)-p;
      R q_=R(1)-q;
      return f11*p_*q+f12*p*q+f21*p_*q_+f22*p*q_;
   }
   else if (method==0){ // pixel representation
      if (p<0.5){
         if (q<0.5) return f21; else return f11;
      }
      else{
         if (q<0.5) return f22; else return f12;
      }
   }
   else {
      cpmerror("Frame", "in2: method not implemented");
      return 1.;
   }
}

V<R> ImagingTools::pixelCenters(R a, R b, Z n)
// The interval [a,b] is divided into n subintervalls of equal length
// and the n centers of these subintervals are returned as a  real vector
// in ascending order
{
   Iv iv(a,b);
   return iv.centers(n);
}

void ImagingTools::find(const V<R>& lattice, R x, R& p, Z& i, Z& j)
// Positioning x within an equidistant 1-dimensinal lattice.
// After function call one of three situations is given:
//
// 1. (normal case)      lattice[i] <= x < lattice[j]
//
// 2. x<lattice[1], i=1, j=2
//
// 3. i=n-1, j=n, lattice[n]<=x   (n=dim(lattice))
//
// p=(x-lattice[i])/(lattice[j]-lattice[i]).
// It is assumed that lattice[j+1]-lattice[j]>0 and independent of j,
// whenever none of the indexes is out of range.
{
   Word loc("ImagingTools::find(...)");
   Z n=lattice.dim();
   if (n==1){p=0.;i=1;j=1;return;}
   R first=lattice[1];
   R last=lattice[n];
   cpmassert(first<last,loc);
   R step=(last-first)/(n-1.); // lattice has n components and n-1
      // gaps between values
   if (x<=first){
      i=1;
      j=2;
   }
   else if (x>=last){
      i=n-1;
      j=n;
   }
   else{
      i=1+cpmtoz((x-first)/step);
         //would be the exact result if the lattice would
         // be exactly equidistant.
         // It is always a numerically perfect solution for
         // the interpolation applicalion since p gets computed
         // accordingly
      j=i+1;
   }
   R lower=lattice[i];
   R upper=lattice[j];
   cpmassert(lower<upper,loc);
   p=(x-lower)/(upper-lower);
      // is negative below first, and larger than 1 above last.
      // This is OK
}

R ImagingTools::response(R x1, R x2, R y1, R y2, R gamma, R x)
// The function x |---> f(x):=response(...,x) has the following
// properties
// f(x1)=y1, f(x2)=y2,
// f(x) is a power function of a linear transform of x.
// 97-4-14: for gamma=-1, f is an exponential function of a linear
// transform of x. This is valuable if the pixel values of the image are
// proportional to log exposure and the response is intended to be a
// power of exposure.
{
   using namespace CpmRootX;
   static const R maxLim=100;
   static const R minLim=0.01;
   static const R hugeNumber = 1.e64;
      // only needed for gamma<0
   if (x2>hugeNumber) x2=hugeNumber;
   if (x<=x1) return y1;
   if (x>=x2) return y2;
   if (x1==x2) return (y1+y2)*0.5;
   R p=(x-x1)/(x2-x1);
   if (gamma==1.){
      return y1+p*(y2-y1);
   }
   else if (gamma==0){
      return y2;
   }
   else if (gamma==0.5){
      return y1+cpmsqrt(p)*(y2-y1);
   }
   else if (gamma>0){
      return y1+cpmpow(p,gamma)*(y2-y1);
   }
   else{ // gamma<0
      if (x1<minLim) x1=minLim;
      if (x2>maxLim) x2=maxLim;
      if (x<=x1) return y1;
      if (x>=x2) return y2;
      R beta=(y2-y1)/(cpmlog(x2)-cpmlog(x1));
      R alpha=y2-beta*cpmlog(x2);
      return alpha+beta*cpmlog(x);
   }
}

R ImagingTools::random(Z i)
{
   return CpmRoot::randomR(i);
}

Z ImagingTools::random(Z i, Z k)
{
   return CpmRoot::ranT<Z>(k,i);
}

////////////////// class Response //////////////////////////////////////

Z Response::levels(20);

R Response::res_(R const& x)const
{
   R y1=f1_(x);
   R y2=trn_(y1);
   R y3=y2*levFac_(y1);
   return ry_.inf()+y3*ry_.abs();
}

Color Response::col_(R const& x)const
{
   R y1=f1_(x);
   R lf=levFac_(y1);
   R phi=cpmpi*y1*1.75; // phi varies between 0 degrees and 315 degrees
   C z;
   z.polar(lf,phi);
   return Color(z,Iv(0.,1.));
}

R Response::operator()(const R& x)const
{
   static R piInv=R(1)/cpmpi;
   static R piHalf=cpmpi*0.5;

   if (mode_==0){
      return
         ImagingTools::response
            (rx_.inf(),rx_.sup(),ry_.inf(),ry_.sup(),gamma_,x);
   }
   else if (mode_==1){ // strange logic, conserved for back-compatibility
      R p=rx_.linear(x,-1,1);
      p=cpmatan(p);
      p=(p+piHalf)*piInv;
      return ry_.inf()+p*ry_.abs();
   }
  /* else if (mode==1){
      R p=rx.linear(x); // p varies from -1 to +1
      p=cpmatan(p)*R(2); // p varies from -pi/2 to +pi/2
      p=(p+piHalf)*piInv;
      return ry.inf()+p*ry.abs(); // p needs to vary from 0 to 1
   }*/
   else if (mode_==2){
      R p=rx_.periodic(x); // varies from 0 to 1
      return ry_.inf()+p*ry_.abs();
   }
   /*else if (mode==3){
      R p=rx.linear(x,-piHalf,piHalf);
      return ry.inf()+(cpmsin(p)+1.)*ry.radius(); // cpmsin(p) varies from
      // -1 to 1, thus cpmsin(p)+1. from 0 to 2
   }*/
   /*else if (mode==3){
      R p=rx.periodic(x); // varies from 0 to 1
      R p2=cpmsin(p*cpmpi);
      return ry.inf()+p2*p2*ry.abs(); // cpmsin(p) varies from
      // -1 to 1, thus cpmsin(p)+1. from 0 to 2
   }*/
   else if (mode_==3){
      R p=rx_.linear(x,-cpmpi,cpmpi); // so, trigonometric
         // functions can be given the argument p
      return ry_.inf()+(cpmsin(p)+1.)*ry_.radius();
      // dark at both ends of the periodicity interval
   }
   else if (mode_==4){
      R y=rx_(x);
      if (y>0) return ry_.inf(); // then y is outside of rx
      else{
         R p=-y/rx_.radius(); // needs to vary from 0 to 1
         return ry_.inf()+p*ry_.abs();
      }
   }
   else if (mode_==5){
      R px=rx_.periodic(x); // varies from 0 to 1
      R p=px*px*cpmpi; // varies from 0 to pi
      p=cpmsin(p); // first half-wave of sine used.
      p*=p; // to make the tangent horizontal even at px=1
         // Consider p as a function of px, thus p(px):
         // Obviously p(0)=p(1)=0. 0<=p(px)<=1 for 0<=px<=1.
         // The maximum of value 1 is attained for some
         // px > 1/2, approximately px = 0.71. So the curve has horizontal
         // tangents at px=0, px=0.71, px=1, and the hill is much steeper
         // to the right side of the summit than to the left side of it.
         // If
      return ry_.inf()+p*ry_.abs();
   }
   else{
      cpmerror("Response::operator()(R): unvalid mode");
      return 0;
   }
}

Color Response::color(R x)const
{
   if (mode_<=5){
      R y=operator()(x);
      return Color(y,y,y);
   }
   else if (mode_==6){
      R yL=0.; // don't change
      R yU=1.; // don't change
      R y=ImagingTools::response(rx_.inf(),rx_.sup(),yL,yU,gamma_,x);
      if (y<yL) return Color(DARKGRAY); //should not happen
      if (y>yU) return Color(WHITE); // should not happen
      Iv rg(yL,yU);
      R phi1=cpmpi*y; // assumes yL=0, yU=1, varies from 0 to pi
      R phi2=phi1*1.75; // 2 pi - pi/4, purple between blue and red
      R dark=cpmsin(phi1*levels);
      dark*=dark;
      dark*=dark;
      R val=R(1)-dark; // darkening by 'Fraunhofer lines'
      C z;
      z.polar(val,phi2);
         // phi2 varies between 0 degrees and 315 degrees
      Color col(z,rg);
      return col;
   }
   else if (mode_==7){
      if (bw_){ R y=res_(x); return Color(y,y,y);}
      else return col_(x);
   }
   else{
      return Color();
   }
}

bool Response::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(rx_);
   cpmp(ry_);
   cpmp(gamma_);
   cpmp(periodic_);
   cpmp(mode_);
   return true;
}

bool Response::scanFrom(istream& str)
{
   cpms(rx_);
   cpms(ry_);
   cpms(gamma_);
   cpms(periodic_);
   cpms(mode_);
   return true;
}

/***************** class Roi Region of Interest ************************/

namespace{
Z mult_pix(Z n, R x)
{
   R s=x*n+0.5;
   cpmassert(s>=0,"mult_pix(Z,R)");
   Z res=cpmtoz(s);
   if (res<1) return 1;
   if (res>n) return n;
   return res;
}
}

Roi::Roi(Z m, Z n, R xL, R xU, R yL, R yU)
{
   Word loc("Roi(Z m, Z n, R xL, R xU, R yL, R yU)");
   cpmassert(xL>=0 && xU<=1,loc);
   cpmassert(yL>=0 && yU<=1,loc);
   cpmassert(xL<xU,loc);
   cpmassert(yL<yU,loc);

   Z j1=mult_pix(n,xL);
   Z j2=mult_pix(n,xU);
   Z i1=mult_pix(m,R(1)-yU);
   Z i2=mult_pix(m,R(1)-yL);
   mc=1+i2-i1;
   nc=1+j2-j1;
   si=i1-1;
   sj=j1-1;
}

Roi::Roi(Z m, Z n, Iv ivx, Iv ivy)
{
   *this=Roi(m,n,ivx[1],ivx[2],ivy[1],ivy[2]);
}

///////// class Rect Rectangular Subareas of the unit square /////////////

Rect::Rect(Iv iv1, Iv iv2)
{
   Iv iv0(0,1);
   iv1=iv1&iv0;
   iv2=iv2&iv0;
   xL=iv1.inf();
   xU=iv1.sup();
   yL=iv2.inf();
   yU=iv2.sup();
}

Rect::Rect(R a, R b, R c, R d)
      // defines a sub-rectangle of the unit-square [0,1]x[0,1].
      // If the arguments are as intended, we have
      // 0<=a<b<=1
      // 0<=c<d<=1.
      // Then the sub-rectangle is simply [a,b]x[c,d]
      // If they are out of range or out of order, they will be brought
      // back in an obvious manner
{
      if (a<=b){
         xL=a;
         xU=b;
      }
      else{
         xL=b;
         xU=a;
      }

      if (c<=d){
         yL=c;
         yU=d;
      }
      else{
         yL=d;
         yU=c;
      }

      if (xL<0) xL=0;
      if (xU>1) xU=1;
      if (yL<0) yL=0;
      if (yU>1) yU=1;
}

Rect::Rect(Z c1x, Z c1y, Z c2x, Z c2y, Z cx, Z cy)
{
   Word loc("Rect(Z c1x, Z c1y, Z c2x, Z c2y, Z cx, Z cy)");
   cpmassert(cx>0,loc);
   cpmassert(cy>0,loc);
   R cxInv=1./cx;
   R cyInv=1./cy;
   R a=R(c1x)*cxInv;
   R b=R(c2x)*cxInv;
   R c=R(1)-R(c1y)*cyInv;
   R d=R(1)-R(c2y)*cyInv;
   *this=Rect(a,b,c,d);
}

namespace{
void auxSub(R a1, R a2, R b1, R b2, R& c1, R& c2)
   // determining subinterval (c1,c2) of (0,1) as subinterval '(b1,b2)'
   // of (a1,a2). The 'subinterval (b1,b2)' is given 'relative to
   // (a1,a2)' i.e. considering (b1,b2) as a unit-interval.
{
   R d=a2-a1;
   c1=a1+b1*d;
   c2=a1+b2*d;
}
}

Rect Rect::sub(const Rect& rec)const
   // returns the rectangle which is formed by taking rec of *this
   // in a way that *this is now considerd a unit square
{
   R x1,x2,y1,y2;
   auxSub(xL,xU,rec.xL,rec.xU,x1,x2);
   auxSub(yL,yU,rec.yL,rec.yU,y1,y2);
   return Rect(x1,x2,y1,y2);
}

bool Rect::prnOn(ostream& str)const
{
   cpmwat;
   cpmp(xL);
   cpmp(xU);
   cpmp(yL);
   cpmp(yU);
   return true;
}

bool Rect::scanFrom(istream& str)
{
   cpms(xL);
   cpms(xU);
   cpms(yL);
   cpms(yU);
   return true;
}

///////////////////// class Mask /////////////////////////////////////////

Mask Mask::invert(void)const
{
   return Mask(-di_,-dj_,w_);
}

R Mask::rms(void)const
{
   Word loc("CpmImaging::Mask::rms(void)");
   R res=0;
   R wTot=0.;
   R r2,dx,dy,wi;
   for (Z i=1;i<=dim();i++){
      dx=di_[i];
      dy=dj_[i];
      r2=dx*dx+dy*dy;
      wi=cpmabs(w_[i]);
      res+=wi*r2;
      wTot+=wi;
   }
   cpmassert(wTot>0.,loc);
   res/=wTot;
   cpmassert(res>=0.,loc);
   return cpmsqrt(res);
}

Mask::Mask(Z n)
{
   if (n==0){ // trivial mask that will make a convolution trivial
      di_=V<Z>(0);
      dj_=V<Z>(0);
      w_=V<R>(0);
      return;
   }
   Z nOdd=(n%2 ? n : n+1);
      // n is odd then the modulo 2 operation % gives 1
   Z nD=nOdd*nOdd;
   cpmassert(nD>0,"Mask::Mask(Z n)");
   R weight=1./nD;

   Va<R> w1(nD,weight);
   w_=w1;

   Va<Z> di1(nD);
   Va<Z> dj1(nD);
   Z i,j,k=1;
   Z nShift=(nOdd+1)/2;
   for (i=1;i<=nOdd;i++){
      for (j=1;j<=nOdd;j++){
         di1[k]=i-nShift;
         dj1[k]=j-nShift;
         k++;
      }
   }
   di_=di1;
   dj_=dj1;
}

Mask Mask::bin3()
{
   V<Z> di{-1,0,1,-1,0,1,-1,0,1};
      // here we need 9 entries, this kind of list
      // constructor is defined for up to 10 entries.
   V<Z> dj{1,1,1,0,0,0,-1,-1,-1};
   V<R> wp{1.,2.,1.,2.,4.,2.,1.,2.,1.};
   Va<R> w(wp);
   R i16=1./16;
   w*=i16;
   return Mask(di,dj,w);
}

Mask::Mask(R rMax, const F<R,R>& f)
{
   Word loc("Mask::Mask(R rMax, const F<R,R>& f)");
   if (rMax<=0.){ // trivial mask that will make a convolution trivial
      di_=V<Z>(0);
      dj_=V<Z>(0);
      w_=V<R>(0);
      return;
   }
   Z i,j,nD=0;
   Z lim=cpmtoz(rMax);
   R rMax2=rMax*rMax;
   R val;
   // counting the contributions needed
   for (i=-lim;i<=lim;i++){
      for (j=-lim;j<=lim;j++){
         val=i*i+j*j;
         if(val<=rMax2) nD++;
      }
   }
   Va<R> w1(nD);
   Va<Z> di1(nD);
   Va<Z> dj1(nD);
   Z k=0;
   R r;
   for (i=-lim;i<=lim;i++){
      for (j=-lim;j<=lim;j++){
         val=i*i+j*j;
         if(val<=rMax2){
            k++;
            r=cpmsqrt(val);
            val=f(r);
            cpmassert(val>=0.,loc);
            w1[k]=val;
            di1[k]=i;
            dj1[k]=j;
         }
      }
   }
   R wSum=0.;
   for (k=1;k<=nD;k++){
      wSum+=w1[k];
   }
   cpmassert(wSum>0,loc);
   wSum=R(1)/wSum;
   w_=w1*wSum;
   di_=di1;
   dj_=dj1;
}

namespace{
   R fAux(const R& r, const R& r0)
   {
      R rRel=r/r0;
      return rRel>1 ? R(0) :R(1)-rRel*rRel;
   }
}

Mask Mask::parabolic(R rMax)
{
#ifdef CPM_Fn
   return Mask(rMax,F1<R,R,R>(rMax)(fAux));
#else
   using std::placeholders::_1;
   using std::bind;
   F<R,R> f(bind(fAux,_1,rMax));
   return Mask(rMax,f);
#endif
}
