//? cpminifilebasapp.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <iomanip>
#include <cpmsystemdependencies.h>
#include <cpmdimdef.h>
#include <cpminifilebasapp.h>
#include <cpmgreg.h>
#include <cpmgraph.h>

using namespace CpmRoot;
using namespace CpmArrays;
using namespace CpmFunctions;
using namespace CpmSystem;

using CpmRootX::RecordHandler;

using CpmGraphics::Frame;
using CpmGraphics::Graph;
using CpmGraphics::Color;

using CpmApplication::IniFileBasedApplication;
using CpmApplication::args_;
using CpmApplication::Task;
using CpmSystem::glue;

//bool CpmApplication::docu=true;
Z CpmApplication::dimApp=2;

//////////////////////////// IniFileBasedApplication ///////////////////

Z CpmApplication::setDimApp( Word const& program, Word const& path)
{
   const Z mL=1;
   Word loc="CpmApplication::setDimApp(";
   loc&=program;
   loc&=")";
   CPM_MA
   V<Word> proIni{program&".ini"}; // only one file name
   Word pathExt=Message::ext(path);
   RecordHandler rch(proIni, pathExt,
      "called from CpmApplication::setDimApp");
   Word sec("selection");
   Z mode_;
   rch.read(sec,"mode",mode_);
   if (mode_!=2 && mode_!=3){
      cpmmessage("mode="&cpm(mode_)&" not useful as space dimension");
   }
   else dimApp=mode_;
   CPM_MZ
   return mode_;
}

IniFileBasedApplication::IniFileBasedApplication(
Word const& program):
size(1),rank(1),
rch(program&".ini",Message::getInpDir(),
   "from IniFileBasedApplication(Word)"),
   // Here already the initialization does substantial work.
   // It loads the data member rch ( for record handler) with the data
   // read from ini-file Message::ext(program&".ini")
inFiles(0),outFiles(0),sel_(0),mode_(0),runName_("xxx"),
appendInFiles_(0),appendOutFiles_(0)
{
   const Z mL=1;
   Word loc="IniFileBasedApplication(";
   loc&=program;
   loc&=")";
   CPM_MA
   programName_=program;
   Word iniFileName(Message::ext(program&".ini"));

   Word sec("selection");
   Z ped=0; // degree of pedantry
   tWaitAnnounceScreen_=4;
   rch.read(sec,"tWaitAnnounceScreen",tWaitAnnounceScreen_,ped);
      // needs not to be there
   ped=1;
   rch.read(sec,"cpmverbose",cpmverbose,ped);
   rch.read(sec,"cpmdbg",cpmdbg,ped);
      // = cpmdbg controls whether cpmassert warns or stops
   CpmTime::Greg date;
   date.now();
   Word datw=date.getDate();
   cpmcerr<<endl<<"date: "<<datw<<endl;
   cpmdata<<endl<<"date: "<<datw<<endl;

//   cpmmessage(" Listing all data from "&iniFileName);
   bool writeComments=false;
//   rch.writeFormatted(cpmcerr,writeComments);
//   cpmmessage(" End of listing all data from "&iniFileName);

   rch.read(sec,"sel",sel_,ped);
   if (sel_<0) Message::setAppRunId(false);
   rch.read(sec,"runName",runName_);
   if (runName_=="auto"){
      runName_=Message::getRunId();
   }
   // coupling the mode parameter to the dimension options set for
// compilation. We do not enforce the definition of CPM_DIMDEF2
// or CPM_DIMDEF3 and create only a warning if it is not defined.
   rch.read(sec,"mode",mode_,ped);
   if (mode_!=2 && mode_!=3){
      cpmmessage("mode="&cpm(mode_)&" not useful as space dimension");
   }
   else{ // in this case mode_ is interpreted as a dimension selector
      dimApp=mode_;
      cpmmessage("mode="&cpm(mode_)&"=dimApp");
#if defined(CPM_DIMDEF2)&&defined(CPM_DIMDEF3)
      cpmmessage("both CPM_DIMDEF2 and CPM_DIMDEF3 were defined");
      cpmmessage("mode and CPM_DIMDEF match");
#elif defined(CPM_DIMDEF2)
      cpmmessage("only CPM_DIMDEF2 was defined");
      if (mode_==3) cpmwarning("mode and CPM_DIMDEF do not match");
#elif defined(CPM_DIMDEF3)
      cpmmessage("only CPM_DIMDEF3 was defined");
      if (mode_==2) cpmwarning("mode and CPM_DIMDEF do not match");
#else
      cpmwarning("Neither CPM_DIMDEF3 nor CPM_DIMDEF2 was defined");
#endif
      if (sel_!=0){
         runName_&="_";
         runName_&=cpmwrite(mode_);
         // mode indicator appended to run name
      }
   }
   if (Message::getAppRunId()){
      runName_&="_";
      runName_&=Message::getRunId();
      Message::setRunId(runName_); // experiment
         // all the sophisticated name extensions are now in the run id
         // and thus can finally appended to cpmcerr for identification.
   }
// consistent creation of input-filenames
// no opening of files !
   sec="input";
   Z i,numberOfFiles=0;
   rch.read(sec,"numberOfFiles",numberOfFiles,ped);
   if (numberOfFiles>0){
      inFiles=V<Word>(numberOfFiles);
      appendInFiles_=V<Z>(numberOfFiles);
      for (i=1;i<=numberOfFiles;++i){
         rch.read(sec,"files"&cpm(i),inFiles[i],ped);
         rch.read(sec,"doc"&cpm(i),appendInFiles_[i],ped);
      }
   }
   Word indictorName="_"&runName_&".txt";
// consistent creation of output-filenames
// no opening of files ! Opening the files for writing is expected to be
// done in function doTheWork() by a statement like
//    OFileStream out1(outFiles[1]);
// Since 2014-4-19 the file names include the directory location setting
// given by Message::outDir which is expected to be set in
// cpmsystemdependencies.h. By code it is made sure that the output files
// will be located in the same directory as cpmcerr.txt and cpmdata.txt.
   sec="output";
   bool succ=rch.read(sec,"numberOfFiles",numberOfFiles,ped);
   Word dio=Message::getOutDir2()&"/";

   if (succ){ // then read outFiles and
      // appendOutFiles_from ini-file
      if (numberOfFiles>0){
         outFiles=V<Word>(numberOfFiles);
         appendOutFiles_=V<Z>(numberOfFiles);
         for (i=1;i<=numberOfFiles;++i){
            rch.read(sec,"files"&cpm(i),outFiles[i],ped);
            outFiles[i]=dio&outFiles[i]&indictorName;
            rch.read(sec,"doc"&cpm(i),appendOutFiles_[i],ped);
            cpmmessage(loc&": outFiles["&cpm(i)&"] = "&outFiles[i]);
         }
      }
   }
   else{ // Then create a outFiles and appendInFiles_
      // automatically such that there is one outFile
      // per inFile and the docu status is the same for
      // in- and out-files
      if (numberOfFiles>0){
         outFiles=V<Word>(numberOfFiles);
         appendOutFiles_=V<Z>(numberOfFiles);
         Word pre("out_");
         for (i=1;i<=numberOfFiles;++i){
           outFiles[i]=pre&inFiles[i].baseName().appBefExt(indictorName);
            appendOutFiles_=appendInFiles_;
            cpmmessage(loc&": outFiles["&cpm(i)&"] = "&outFiles[i]&
               " !auto-created !");
         }
      }
   }
   // creating (for output) and opening of these files has to occur in
   // function doTheWork. In function documentTheWork, the output files
   // will be tried to open. If this is not possible a message on cpmcerr
   // will document this. The reason can be that function doTheWork did
   // not create the file.
   if (cpmverbose>=mL && rank==1&&!cpmsilent()){
      cpmdata<<endl<<"Input values as read from file:"<<endl;
      rch.getRecord().writeFormatted(cpmdata,writeComments);
   }
   tStart_=cpmtime();
   CPM_MZ
}

namespace{
   Word chain(V<Word> const& vw)
      // putting words together with a separating blank
   {
      Word res;
      for (Z i=vw.b();i<vw.e();++i) res&=(vw[i]&" ");
      res&=vw.last();
      return res;
   }
}

void IniFileBasedApplication::announceTheWork()
{
   Z mL=1;
   Word loc("IniFileBasedApplication::announceTheWork()");
   CPM_MA
   if (tWaitAnnounceScreen_<=0){
      CPM_MZ
      return;
   }
// defining colors
   Color cBack(Z(0),Z(0),Z(60));
      // background color
   Color cProg(Z(255),Z(255),Z(255));
      // color for program title and the last line
   Color cTopics(Z(255),Z(255),Z(0));
      // color for the topics under consideration
   Color cSources(Z(50),Z(50),Z(120));
      // color for the text concerning sources
   Color cInfo(Z(100),Z(100),Z(240));
      // color for the text concerning info

   Word file=programName_;
      // name of the ini-file but without file extension
      // This name may be overridden by reading a name from the
      // ini-file

   Word beg0="        ";
   Word wdef=beg0&" ... ";
   V<Word> line1(1,wdef);
   V<Word> line2(1,wdef);
   V<Word> line3(1,wdef);
   V<Word> line4(1,wdef);
   V<Word> line5(1,wdef);
   V<Word> line6(1,wdef);
   V<Word> line7(1,wdef);
   V<Word> line8(1,wdef);

   Word sec="about";
// optionally overriding preset values
   cpmrhf(file);
   cpmrhf(line1);
   cpmrhf(line2);
   cpmrhf(line3);
   cpmrhf(line4);
   cpmrhf(line5);
   cpmrhf(line6);
   cpmrhf(line7);
   cpmrhf(line8);

// The topic lines enter primarily as lists of words
// which will be combined into a single Word containing spaces.
   Word beg1="Topics: ";
   V<Word> topics(0);
   topics<<beg1+chain(line1);
   topics<<beg0+chain(line2);
   topics<<beg0+chain(line3);
   topics<<beg0+chain(line4);
   topics<<beg0+chain(line5);
   topics<<beg0+chain(line6);
   topics<<beg0+chain(line7);
   topics<<beg0+chain(line8);

   V<Word> lines;
   V<Color> colors;
   Word tit=CpmApplication::title_;

   lines<<("Program: "&tit);
   colors<<cProg;

   lines<<"";
   colors<<cProg;

   lines<<"Editing the input-file "+file+
      " allows controlling the program.";
   colors<<cInfo;

   lines<<"See this file and "+tit+"_readme.txt"+" for information.";
   colors<<cInfo;

   lines<<"";
   colors<<cProg;

   for (Z i=topics.b();i<=topics.e();++i){
      lines<<topics[i];
      colors<<cTopics;
   }
   lines<<"";
   colors<<cProg;

   lines<<
      "This program is based on the C+- class system by Ulrich Mutze,";
   colors<<cSources;

   lines<<
      "code modified from Numerical Recipes in C by William H. Press,";
   colors<<cSources;

   lines<<
      "Saul A. Teukolsky, William T. Vetterling, and Brian P. Flannery,";
   colors<<cSources;

   lines<<"the OpenGL Utility Toolkit (GLUT) by Mark J. Kilgard,";
   colors<<cSources;

   lines<<"and the C++ programming language by Bjarne Stroustrup.";
   colors<<cSources;

   lines<<"";
   colors<<cProg;

   lines<<"The program will start when the status bar shows tLeft = 0.";
   colors<<cProg;

   Frame fr(cBack);
   fr.setText(lines,colors);
      // placing lines of text reasonably into
      // the available space (adjusts font size to needs)
   bool writeOnShow=(cpmverbose>1);
   fr.vis(writeOnShow);
   tWaitWorkScreen_=4;
   using CpmApplication::args_;
   if (args_.dim()>=4) tWaitAnnounceScreen_=args_[4].toR();
   if (args_.dim()>=6) tWaitWorkScreen_=args_[6].toR();
      // from the command line or the ini-file one may have
      // a different request that, then, has priority
   cpmwait(tWaitAnnounceScreen_,2);
      // show the remaining time on pane 2
   CPM_MZ
}

void IniFileBasedApplication::documentTheWork()
{
   Z mL=1;
   Word loc="IniFileBasedApplication::documentTheWork()";
   CPM_MA
   tStop_=cpmtime();
   R tRun=tStop_-tStart_;
   using CpmRootX::toString;
   ostringstream ost;
   Z i,nIn=inFiles.dim();
   Z nOut=outFiles.dim();
   CpmTime::Greg date;
   date.now();
   Word dn=date.getDate(false);
   ost<<"Date: "<<dn;
   ost<<endl<<"run time in seconds: "<<tRun<<endl;
   ost<<"Cpm documentation file for selection "<<sel_
      <<" of program "<<programName_
      <<endl<<" and run "<<runName_<<endl
      <<"What follows is obtained by concatenating all input files"
      <<" and all output files"<<endl;
   Word inFiles0=rch.getFileName();
   IFileStream iTemp1(inFiles0);
   ost<<endl<<("Input file "&inFiles0&": ")<<endl<<toString(iTemp1());
   for (i=1;i<=nIn;i++){
      ost<<endl<<"Input file #"<<i<<": "<<inFiles[i]<<endl;
      if (appendInFiles_[i]==0){
            ost<<" not shown";
      }
      else{
         IFileStream iTemp2(inFiles[i]);
         ost<<toString(iTemp2());
      }
   }
   if(nOut>0){
      for (i=1;i<=nOut;i++){
         ost<<endl<<"Output file #"<<i<<": "<<outFiles[i];
         if (appendOutFiles_[i]==0){
            ost<<" not shown";
         }
         else{
            bool safe=false;
            IFileStream oTemp(outFiles[i],safe);
               // IFile is OK here since we'll read
               // the outfiles which have been created (or not)
            if (!oTemp()){
               Word mes=loc;
               mes&=": unable to open "&outFiles[i]&
                  ";\n was probably not created in function doTheWork";
               cpmmessage(mes);
               ost<<" not created";
            }
            else{
               ost<<endl<<toString(oTemp());
            }
         }
      }
   }
   Word docuFileName=
      Message::extOut(programName_&"_"&runName_&"_docu.txt");
   OFileStream out(docuFileName);
   out()<<ost.str();
   cpmdata<<endl<<endl<<"// Full content of rch:"<<endl;
   bool wrtCom=(sel_==0 ? false : true);
   rch.getRecord().writeFormatted(cpmdata,wrtCom);
   out()<<endl<<endl<<"// Full content of rch:"<<endl;
   rch.getRecord().writeFormatted(out(),wrtCom);
   // added 2004-09-14 since with the new hierarchical feature of
   // ini-files a mere copy of the top-level ini-file does not allow
   // to trace all data that made the program run as it did.
   // So we include a concise (reordered) record of all data here.
   // Even the comments are written in order not to loose comments
   // explaining newly introduced quantities.
   CpmApplication::inspectCpmdefinitions();
   CpmApplication::inspectCpmsystemdependencies();
  // CpmSystem::docRun(false);
      // already called from cpminifilebasapp.cpp. Woud result in an boost
      // error if would also be called from here
   CPM_MZ
}

void IniFileBasedApplication::run()
{
   cpmurgent("IniFileBasedApplication::run() started");
   cpmstatus(programName_&" started",4); //
   announceTheWork();
   R ti=cpmtime();
   doTheWork();
   R tf=cpmtime();
   R tfi=tf-ti;
   cpmcerr<<endl<<"execution time for doTheWork() was "<<tfi<<" s"<<endl;
   bool writeComments=(sel_>0 ? true : false);
   if (Message::getAppRunId() && rank==1&&!cpmsilent()){
      Word leanDocuFileName=
      Message::extOut(programName_&"_"&runName_&"_.ini");
      OFileStream ldf(leanDocuFileName);
      rch.getRecord().writeFormatted(ldf(),writeComments);
      if (sel_>0) documentTheWork(); // also docu of results
   }
   cpmurgent("IniFileBasedApplication::run() done");
      // to get timing data also for cpmverbose=0. For comparing
      // efficiency of compilers it is wise reduce messaging to
      // a minimum, but timing is of impotance then
   cpmstatus(programName_&" done",4); //
   cpmwait(15.,2); // experiment for inspecting final console content.
}

void IniFileBasedApplication::setModeFromInFiles()
   // In some programs based on IniFileBasedApplication the inFiles
   // are assumed to be the outFiles of other programs also based
   // on IniFileBasedApplication. These outfiles have the mode
   // parameter normally appended, and thus allow the mode-value
   // which fits the nature of these files to be read from
   // the file names of the inFiles under consideration. This
   // mechanism (which changes the values of mode and runName)
   // will be executed by calling the present function
{
   Word loc("IniFileBasedApplication::setModeFromInFiles()");
   Word t2="2.txt";
   Word t3="3.txt";
   Z tL=5; // tL==t2.dim()==t3.dim()
   Z d=inFiles.dim();
   cpmassert(d>0,loc);
   Word tail=inFiles[1].tail(tL);
   Z res{0};
   bool b(true);
   if (tail==t2) res=2;
   else if(tail==t3) res=3;
   else b=false;
   for (Z i=2;i<=d;i++){
      if (inFiles[i].tail(tL)!=tail) b=false;
   }
   if (b){
      mode_=res;
      Word mo=(mode_==2 ? "2" : "3");
      Word rnt=runName_.tail(1);
      if (rnt=="2" || rnt=="3"){
         // if runName ends in 2 or 3, it should get the end
         // which fits mode
         if (rnt!=mo) runName_=runName_.cut(1)&mo;
      }
   }
}

namespace{

bool ini_args_(V<Word> const& fl)
{
   Z mL=1;
   Word loc("CpmApplication::ini_args_(V<Word>)");
   RecordHandler rch(fl,"",loc);
   if (!rch.isVal()){
      cpmwarning(loc&
         ": RecordHandler not valid, commandline not extended");
      CPM_MZ
      return false;
      // Thus no action if file fn does not exist
   }
   V<Word> args=rch.comLine();
   args_=CpmRootX::combine(args_,args);
   CPM_MZ
   return args_.dim()>0;
}

bool ini_args_(Word const& fl)
{
   Z mL=1;
   Word loc("CpmApplication::ini_args_(Word)");
   RecordHandler rch(fl,"","CpmApplication::ini_args_("&fl&")");
   if (!rch.isVal()){
      cpmwarning(loc&
         ": RecordHandler not valid, commandline not extended");
      CPM_MZ
      return false;
      // Thus no action if file fn does not exist
   }
   V<Word> args=rch.comLine();
   args_=CpmRootX::combine(args_,args);
   CPM_MZ
   return args_.dim()>0;
}

Z menFun(Word const& pn, V< F<Word,Z> > const& appList)
{
   Word loc("menFun("&pn&",V< F<Word,Z> >)");
   Z mL=1;
   CPM_MA
 //  Word iniFile=pn&".ini";
   Word iniFile=glue(Message::getInpDir(),pn&".ini");
   cpmmessage("iniFile = "&iniFile);
   Word dummy;
   Z i;
   bool succ=ini_args_(iniFile);
   if (!succ) cpmerror("ini-file "&iniFile
      &" not found");
   Z sel=1;
   R tWait=1;
   Z na=args_.dim();
   cpmmessage(mL,"args_.dim()="&cpm(na));
   if (na>=2) sel=args_[2].toZ();
   if (na>=3) cpmverbose=args_[3].toZ();
   if (na>=4) tWait=args_[4].toR();
   cpmmessage("sel="&cpmwrite(sel));
   cpmmessage("cpmverbose="&cpmwrite(cpmverbose));
   Z nn=appList.dim();
   if (sel<1) cpmerror("main_(): sel<1");
   else if (sel>nn){ // doing all
      cpmmessage("sel="&cpm(sel)&
         " asks for doing the whole list");
      for (i=1;i<=nn;++i){
         cpmstatus("menu item "&cpm(i)&" started",4);
         appList[i](dummy);
         cpmstatus("menu item "&cpm(i)&" done",4);
      }
   }
   else{ // doing only the selected item
      cpmmessage("sel="&cpm(sel)&
         " asks for doing a single list component");
      cpmstatus("menu item "&cpm(sel)&" started",4);
      appList[sel](dummy);
      cpmstatus("menu item "&cpm(sel)&" done",4);
      // this calls the selected program
   }
   cpmwait(tWait);
   CPM_MZ
   return sel;
}

Z subMenFun(Word const& dummy, Word const& pn,  F<Word,Z>  const& f)
{
   Word loc("subMenFun("&pn&", F<Word,Z>)");
   Z mL=1;
   CPM_MA
   dummy; // not needed as the name says
// trying to find data
   Word iniFile=glue(Message::getInpDir(),pn&".ini");
   RecordHandler rch(iniFile,"","subMenFun(...)");
   bool succ=rch.isVal();
   if (!succ){
      cpmerror(loc&": no valid ini-file found");
      return 0; // will never be reached
   }
   else{
      cpmmessage(loc&": got valid data from "&rch.getFileName());
   }
// extracting task-related file names
   Word sec("tasks");
   Z begin=1,end=1;
   rch.read(sec,"begin",begin);
   rch.read(sec,"end",end);
   Z ped=0;
   V<Z> ignore(0);
   rch.read(sec,"ignore",ignore,ped);
   Z di=ignore.dim();
   ped=2;
   succ=false;
   V<Word> tasks;
   V<Word> varTask;
   succ=rch.read(sec,"tasks01",tasks,ped);
   ped=1; // further tasks may not be there
// maximum of 20 task groups prepared; since each task group
// may have as many tasks as one may write names in a line of the
// ini-file, this is not really a restriction. However, having only
// a small constant number, say 3, in a line, limits the total
// number (to 60 for that number, which is probably more than ever
// needed.
   if (succ){
      succ=rch.read(sec,"tasks02",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks03",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks04",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks05",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks06",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks07",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks08",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks09",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks10",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks11",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks12",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks13",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks14",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks15",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks16",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks16",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks17",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks18",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks19",varTask,ped);
      if (succ) tasks&=varTask;
   }
   if (succ){
      succ=rch.read(sec,"tasks20",varTask,ped);
      if (succ) tasks&=varTask;
   }
   Z nTasks=tasks.dim();
   cpmassert(begin>=1,loc);
   cpmassert(begin<=nTasks,loc);
   cpmassert(end>=1,loc);
   cpmassert(end<=nTasks,loc);
   sec="control";
   R tWait=0;
   rch.read(sec,"tWait",tWait);
   Z i,sum=0;
   for (i=begin;i<=end;++i){
      bool vi=true;
      if (di>0){
         for (Z j=1;j<=di;++j){
            if (i==ignore[j]){
               vi=false;
               break;
            }
         }
      }
      if (vi){
         Word iniFileForTask=tasks[i];
         cpmmessage(iniFileForTask&" considered as task");
         sum+=f(iniFileForTask);
            // the tasks determine the name of the ini-file which
            // will deliver data
         cpmwait(tWait);
      }
   }
   CPM_MZ
   return sum;
}

}// namespace

Task Task::subMen(Word const& name, F<Word,Z> const& f)
{
#ifdef CPM_Fn
   F<Word,Z> f1=F2<Word,Word,F<Word,Z>,Z>(name,f)(subMenFun);
#else
   F<Word,Z> f1(bind(subMenFun,_1,name,f));
#endif
   return Task(name,f1);
}

Task Task::men(Word const& name, V<Task> const& tasks)
// That this works shows the power of the function type
{
   Z td=tasks.dim();
   V< F<Word,Z> > fs(td);
   for (Z i=1;i<=td;++i) fs[i]=tasks[i].get_f();
#ifdef CPM_Fn
   F<Word,Z> f=F1< Word, V< F<Word,Z> >, Z >(fs)(menFun);
#else
   F<Word,Z> f(bind(menFun,_1,fs));
#endif
   return Task(name,f);
}

bool Task::prnOn(ostream& str)const
{
   CpmRoot::writeTitle("Task",str);
   cpmp(name_);
   cpmp(f_);
   return true;
}

bool Task::scanFrom(istream& str)
{
   cpms(name_);
   cpms(f_);
   return true;
}
////////////////////////////////////////////////////////////////////////

Z CpmApplication::menueGenerator(Word const& title, const V<Word>& items,
 const V< F<Z,Z> >& f, Z lang )
{
   Z i;
   Word menue1, menue2, askForEnter, askForExit, choice1, choice2,
    choice3;
   if (lang==1){
      menue1="Folgende Moeglichkeiten stehen zur Auswahl:";
      menue2="Nach dieser Berechnung stehen folgende Moeglichkeiten zur\
       Auswahl:";
      askForEnter="Eingabe der Wahlnummer:  ";
      askForExit="Programm beenden";
      choice1="Programm\
       beenden........................................... 0";
      choice2="Rueckkehr zum\
       Hauptmenue................................... 1";
      choice3="Rueckkehr zum letzten\
       Menuepunkt........................... 2";
   }
   else {
      menue1="Main Menue: the following selections are possible";
      menue2="Continuation Menue: the following selections are possible";
      askForEnter="Enter selection:  ";
      askForExit="EXIT from program";
      choice1="EXIT from\
       program.......................................... 0";
      choice2="Return to Main\
       Menue....................................... 1";
      choice3="Return to begin of current\
       item............................ 2";
   }

   V<Word> itemsNew=items&askForExit;
   Z nMenue=itemsNew.dim();
   Vo<Z> sizes(nMenue);
   for (i=1;i<=nMenue;i++) sizes[i]=itemsNew[i].dim();
   Vo<Z> orderedSizes=sizes.decreasingCopy();
   Z maxSize=orderedSizes[1];
   maxSize+=2;

MAINPOINT: cout<<endl<<title;
   cout<<endl<<menue1;
   Z fillsize;
   for (i=1;i<=nMenue;i++){
      cout<<endl;
      cout<<itemsNew[i];
      fillsize=maxSize-sizes[i];
      for (Z j=1;j<=fillsize;j++) cout<<".";
      cout<<" "<<i;
   }
   cout<<endl<<askForEnter;
   Z sel; cin>>sel;
   if (sel==nMenue) goto END;
SUBPOINT:
   f[sel](1);
   cout<<endl<<menue2;
   cout<<endl<<choice1;
   cout<<endl<<choice2;
   cout<<endl<<choice3;
   cout<<endl<<askForEnter;
   Z sel2; cin>>sel2;
   switch(sel2){
      case 0:  goto END;
      case 1:  goto MAINPOINT;
      case 2:  goto SUBPOINT;
      default: goto END;
   }
END:
   return 1;
}
