//? cpmmacros.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_MACROS_H_
#define CPM_MACROS_H_
/*

   Description: macros for uniform implementation of
      functions. Too often I wrote the starting cpmmessage
      and the closing one in function blocks. The abbreviation by macros
      is not impressive in this case. For the R-macro the savings are
      more convincing.

*/

//////////////////////// CPM_MAZ //////////////////////////////////////

// M stands for messaging
//              -
// A for beginning, Z for end (from the
// role of A and Z in the alphabet)

// needs
// Z mL=...;
// Word loc(".....");
// see the example following CPM_RAZ

#define CPM_MA\
   cpmmessage(mL,loc&" started");

#define CPM_MZ\
   cpmmessage(mL,loc&" done");

//////////////////////// CPM_RAZ //////////////////////////////////////

// R stands for report
//              -
// reports the computation time for the code between CPM_RA and
// CPM_RZ. If loadMes (load measure) is assigned a positive value
// also the computation time divided by that quantity is returned
// under the name of 'normalized computation time'

// needs
// Word loc(".....");

#define CPM_RA\
   R tcStart=cpmtime();\
   R loadMes=0;

#define CPM_RZ\
   R tcFinal=cpmtime();\
   R tcTotal=tcFinal-tcStart;\
   std::ostringstream ost;\
   ost<<std::endl<<"Performance data for "<<loc.toStr()<<std::endl;\
   ost<<"computing time ="<<tcTotal;\
   if (loadMes>0){\
      R tcNorm=tcTotal/loadMes;\
      ost<<std::endl<<"normalized computing time="<<tcNorm;\
   }\
   cpmmessage(1,Word(ost.str()),-1);
      // not writing to the status bar due to '-1'

///////////////////// usage of CPM_MAZ and CPM_RAZ //////////////////////
/*
   a typical usage is

  void aFunction()
  {
      Z mL=2;
      Word loc("aFunction()");
      CPM_MA
      CPM_RA
      ...
      loadMes=...;
      ...
      CPM_RZ
      CPM_MZ
   }

   for a function that notifies start, end, computation time,
   and an efficiency measure on cpmcerr.txt. Notice that A and Z
   resemble opening and closing brackets.
*/
#endif
