//? cpmoperator.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmoperator.h>

using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::Operator;
using CpmQuantumMechanics::Opr;
using CpmQuantumMechanics::Observable;
using CpmQuantumMechanics::TrjCon;
using CpmQuantumMechanics::TrjB;
#ifdef CPM_Fn
using CpmFunctions::F1;
using CpmFunctions::F2;
#endif
using CpmFunctions::R_Func;
using CpmFunctions::R2_Func;

//////////////////////// class Operator //////////////////////////

namespace{

   State xsrobs(X2<State,R> const& xsr, Observable const& ob)
   {
      return ob*xsr.c1();
   }
}

Operator::Operator(Observable const& ob):d_(ob.dim())
{
#ifdef CPM_Fn
   f_=F1<XSR,Observable,State>(ob)(xsrobs);
#else
   f_=F<XSR,State>(bind(xsrobs,_1,ob));
#endif
}

namespace{
   State sinParHamFunc(X2<State,R> const& st, R2_Func const& pot,
      R const& lambda)
   { return st.c1().sinParHam(pot,st.c2(),lambda);}

   State freeParFunc(State const& st)
   { return st.freeParHam();}
}

Operator Operator::sinParHam(R2_Func const& pot,Z d, R lambda)
{
   return
#ifdef CPM_Fn
   Operator(d,F2<XSR,R2_Func,R,State>(pot,lambda)(sinParHamFunc));
#else
   Operator(d,F<XSR,State>(bind(sinParHamFunc,_1,pot,lambda)));
#endif
}

Operator Operator::freePar(Z d)
{
   return Operator(d,F<State,State>(freeParFunc));
}

Operator::Operator(Z d, F<State,State> const& f):d_(d)
{
#ifdef CPM_Fn
   f_=F1<XSR,F<State,State>,State>(f)(conInTime);
#else
   f_=F<XSR,State>(bind(conInTime,_1,f));
#endif
}

Operator Operator::operator*(Operator const& op)const
{
   return
#ifdef CPM_Fn
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(mul2));
#else
   Operator(d_,F<XSR,State>(bind(mul2,_1,*this,op)));
#endif
}

Operator Operator::operator+(Operator const& op)const
{
   return
#ifdef CPM_Fn
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(add2));
#else
   Operator(d_,F<XSR,State>(bind(add2,_1,*this,op)));
#endif
}

Operator Operator::operator-(Operator const& op)const
{
   return
#ifdef CPM_Fn
   Operator(d_,F2<XSR,Operator,Operator,State>(*this,op)(dif2));
#else
   Operator(d_,F<XSR,State>(bind(dif2,_1,*this,op)));
#endif
}

Operator Operator::operator*(C z)const
{
   return
#ifdef CPM_Fn
   Operator(d_,F2<XSR,Operator,C,State>(*this,z)(mulC));
#else
   Operator(d_,F<XSR,State>(bind(mulC,_1,*this,z)));
#endif
}

Operator Operator::operator*(R r)const
{
   return
#ifdef CPM_Fn
   Operator(d_,F2<XSR,Operator,R,State>(*this,r)(mulR));
#else
   Operator(d_,F<XSR,State>(bind(mulR,_1,*this,r)));
#endif
}

Operator& Operator::operator+=(Operator const& fop)
{
   return *this=operator+(fop);
}

Operator& Operator::operator-=(Operator const& fop)
{
   return *this=operator-(fop);
}

Operator& Operator::operator*=(C z)
{
   return *this=operator*(z);
}

Operator& Operator::operator*=(R r)
{
   return *this=operator*(r);
}

Observable Operator::toObservable(R t)const
{
   Observable res(d_);
   for (Z j=1;j<=d_;++j){
      State cbj= actOn(State(d_,j),t); // cb: canonical basis
      for (Z i=1;i<=d_;++i){
         res.setLean(i,j,cbj.cui(i));
      }
   }
   return res;
}

//Observable Opr::toObservable()const
//{
//   Z mL=1;
//   static Word loc("Opr::toObservable");
//   CPM_MA
//   Observable res(d_);
//   for (Z i=1;i<=d_;++i){
//      State cbi= f_(State(d_,i)); // cb : canonical basi
//      cout<<"||cbi||="<<cbi.abs()<<endl;
//      res.c().cui(i)=cbi.cv();
//   }
//   cout<<"Opr::toObservable()"<<endl;
//   cout<<"res = "<<res<<endl;
//   CPM_MZ
//   return res;
//}

Observable Opr::toObservable()const
{
   Z mL=1;
   static Word loc("Opr::toObservable");
   CPM_MA
   Observable res(d_);
   for (Z j=1;j<=d_;++j){
      State cbj= f_(State(d_,j)); // cb : canonical basi
      for (Z i=1;i<=d_;++i){
         C cji=cbj.cui(i);
         res.setLean(i,j,cji);
      }
   }
   CPM_MZ
   return res;
}

//////////////////////// class TrjTD ///////////////////////////////
/*
void TrjTD::adjustTD_()
{
   State opx=op_.actOn(x_,t_);
   enr_=(x_|opx);
   v_=opx*QMBase::get_ihDiracInv();
   R n1=x_.norm();
   n_1_=n1-1;
   R n2=v_.norm();
   dt_=QMBase::step(n1,n2);
   text_="Evolution by the Time Dependent Direct Midpoint Integrator";
}

C TrjTD::eTot()const
{
   return (x_|v_)*QMBase::get_ihDirac();
}

TrjTD::TrjTD(const Operator& a, const State& x, TrjCon const& tc):
TrjB(x,tc),op_(a)
{
   adjustTD_();
}

void TrjTD::step_(R h)
{
   static C fac=QMBase::get_ihDiracInv()*QMBase::get_ihDiracInv();

   if (autoStep()){ // due to time dependence of
      // the generator, the proposed time step will vary with time
      // and has to be computed for each step (rather cheaply)
      R n1=x_.norm();
      R n2=v_.norm();
      dt_=QMBase::step(n1,n2);
   }
   else dt_=h;

   R tau=0.5*dt_;
   R ti=t_+tau;
   State xi=x_+v_*tau;
   State yi=op_.actOn(xi,ti);
   State dv=op_.actOn(yi,ti)*(fac*dt_);
   x_+=(v_*dt_+dv*tau);
   v_+=dv;
   t_+=dt_;
   stepCon_+=1;
   norIfEnf_();
   enr_=eTot();
}

bool TrjTD::prnOn(ostream& str)const
{
   cpmwt("TrjTD");
   if (! TrjB::prnOn(str)) return false;
//   cpmp(op_);
   cpmp(v_);
   return true;
}

bool TrjTD::scanFrom(istream& str)
{
   if (! TrjB::scanFrom(str)) return false;
//   cpms(op_);
   cpms(v_);
   return true;
}
*/
