//? cpmrmateig.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#include <cpmrmateig.h>
#include <cpmrmatrix.h>
#include <cpmvectormore.h>
#include <cpmgraph.h>
#include <cpmsystemdependencies.h>

using namespace CpmRoot;
using namespace CpmRootX;
using namespace CpmSystem;
using namespace CpmArrays;
using CpmGraphics::Graph;

namespace{ // anonymous namespace

////////////// functions modified from Press et al. ///////////

void tred2(R_Matrix& a, R_Vector& d, R_Vector& e)
// all arguments may be changed
// Implementation in which all literals are typed as _R or _Z.
// Looks quite practical.
{
   const Z mL=4_Z;
   Z n=a.dim1();
   Z n2=a.dim2();
   cpmassert(n==n2,"tred2(...)");
   R_Vector d_(n);
   d=d_;
   R_Vector e_(n);
   e=e_;

   Z l,k,j,i;
   R scale,hh,h,g,f;
   R nInv=1_R/n;
   for (i=n;i>=2_Z;i--) {
      if (cpmverbose>=mL)
         cpmprogress("tred2()-loop 1",R(n-i)*nInv);
      l=i-1_Z;
      h=scale=0_R;
      if (l > 1_Z) {
         for (k=1_Z;k<=l;k++)
            scale += cpmabs(a[i][k]);
         if (scale == 0_R)
            e[i]=a[i][l];
         else {
            for (k=1_Z;k<=l;k++) {
               a[i][k] /= scale;
               h += a[i][k]*a[i][k];
            }
            f=a[i][l];
            g = f>0_R ? -cpmsqrt(h) : cpmsqrt(h);
            e[i]=scale*g;
            h -= f*g;
            a[i][l]=f-g;
            f=0_R;
            for (j=1;j<=l;j++) {
            // Next statement can be omitted if eigenvectors not wanted
               a[j][i]=a[i][j]/h;
               g=0_R;
               for (k=1_Z;k<=j;k++)
                  g += a[j][k]*a[i][k];
               for (k=j+1_Z;k<=l;k++)
                  g += a[k][j]*a[i][k];
               e[j]=g/h;
               f += e[j]*a[i][j];
            }
            hh=f/(h+h);
            for (j=1_Z;j<=l;j++) {
               f=a[i][j];
               e[j]=g=e[j]-hh*f;
               for (k=1;k<=j;k++)
                  a[j][k] -= (f*e[k]+g*a[i][k]);
            }
         }
      } else
         e[i]=a[i][l];
      d[i]=h;
   }
   // Next statement can be omitted if eigenvectors not wanted
   d[1]=0_R;
   e[1]=0_R;
   // Contents of this loop can be omitted if eigenvectors not
   //      wanted except for statement d[i]=a[i][i];
   for (i=1_Z;i<=n;i++) {
      if (cpmverbose>=mL)
         cpmprogress("tred2()-loop 2",R(i-1)*nInv);
      l=i-1_Z;
      if (d[i]!=0_R) {
         for (j=1_Z;j<=l;j++) {
            g=0_R;
            for (k=1_Z;k<=l;k++)
               g += a[i][k]*a[k][j];
            for (k=1_Z;k<=l;k++)
               a[k][j] -= g*a[k][i];
         }
      }
      d[i]=a[i][i];
      a[i][i]=1_R;
      for (j=1_Z;j<=l;j++) a[j][i]=a[i][j]=0_R;
   }
}

R sign(R a, R b){ return (b<0_R ? -cpmabs(a) : cpmabs(a));}

void tqli(R_Vector& d, R_Vector& e, R_Matrix& z, Z iterMax)
// d,e,z should be on input equal to d,e,a of tred2 on output
{
   const Z mL=4;
   Z n=d.dim();
   Z n2=e.dim();
   cpmassert(n==n2,"tqli(...)");
  // cout<<"iterMax = "<<iterMax<<endl;

   Z m,l,iter,i,k;
   Z iterSup=0_Z;
   R iterMaxInv=1_R/iterMax;
   R s,r,p,g,f,dd,c,b;

   for (i=2;i<=n;i++) e[i-1]=e[i];
   e[n]=0_R;
   R nInv=1_R/n;
   for (l=1;l<=n;l++) {
      if (cpmverbose>=mL){
         cpmprogress("tqli()-loop",R(i-1)*nInv);
      }
      iter=0_Z;
      do {
         for (m=l;m<=n-1;m++) {
            dd=cpmabs(d[m])+cpmabs(d[m+1]);
            if ((cpmabs(e[m])+dd) == dd) break;
         }
         if (m != l) {
            if (iter>iterSup){
               iterSup=iter;
               if (cpmverbose>=mL)
                  cpmprogress("iterSup",R(iter)*iterMaxInv,2);
            }
            if (iter++ == iterMax)
               cpmerror("Too many iterations in tqli()");
            g=(d[l+1]-d[l])/(e[l]*2_R);
            r=cpmsqrt((g*g)+1_R);
            g=d[m]-d[l]+e[l]/(g+sign(r,g));
            s=c=1_R;
            p=0_R;
            for (i=m-1;i>=l;i--) {
               f=s*e[i];
               b=c*e[i];
               if (cpmabs(f) >= cpmabs(g)) {
                  c=g/f;
                  r=cpmsqrt((c*c)+1_R);
                  e[i+1]=f*r;
                  c *= (s=1_R/r);
               } else {
                  s=f/g;
                  r=cpmsqrt((s*s)+1_R);
                  e[i+1]=g*r;
                  s *= (c=1_R/r);
               }
               g=d[i+1]-p;
               r=(d[i]-g)*s+c*b*2_R;
               p=s*r;
               d[i+1]=g+p;
               g=c*r-b;
               // Next loop can be omitted if eigenvectors not wanted
               for (k=1;k<=n;k++) {
                  f=z[k][i+1];
                  z[k][i+1]=s*z[k][i]+c*f;
                  z[k][i]=c*z[k][i]-s*f;
               }
            }
            d[l]=d[l]-p;
            e[l]=g;
            e[m]=0_R;
         }
      } while (m != l);
   }
}

///////////// end functions modified from Press et al./////////

} // anonymous namespace
#if defined(CPM_USE_EIGEN) && !defined(CPM_QUAD)
   R_Vector CpmArrays::eigen(R_Matrix const& a, R_Matrix & eig,
   bool test, Z iterMax)
{
   using namespace Eigen;
   const Z mL=2;
   static Word loc="CpmArrays::eigen(...)";
   CPM_MA
   Z i,j,k,l;
   Z n=a.dim1(),n1=a.dim2();
   if (n!=n1) cpmerror("CpmArrays::eigen(...): dimension mismatch");
   Matrix<R,Dynamic,Dynamic> ea(n,n);
   for (i=1,k=0;i<=n;++i,++k) for (j=1,l=0;j<=n;++j,++l)
      ea(k,l) = a[i][j];
   SelfAdjointEigenSolver<Matrix<R,Dynamic,Dynamic> > ses(ea);
   Matrix<R,Dynamic,1> val=ses.eigenvalues();
   Matrix<R,Dynamic,Dynamic> vec=ses.eigenvectors();
   for (i=1,k=0;i<=n;++i,++k) for (j=1,l=0;j<=n;++j,++l)
      eig[i][j]=vec(k,l);
   R_Vector d(n);
   for (i=1,k=0;i<=n;++i,++k)
      d[i] = val(k);
   V<Z> p=(-d).permutationForIncreasingOrder();
   d=d.permute(p);
 //  d*=eSup; // corrects for the scaling of the matrix

   eig=eig.permuteColumns(p);
   eig=~eig;
   for (i=1;i<=n;i++){
      Iv ise=eig[i].infSup();
      R vmin=ise.inf(),vmax=ise.sup();
      if (cpmabs(vmin)>cpmabs(vmax)) eig[i]*=(-1.);
   }
   if (test){
      R tWait=1;
      cpmmessage("CpmArrays::eigen(...) test started");
      R aN=a.abs();
      cpmmessage("The norm of the matrix under consideration is  "&
         cpm(aN));
      R err, errCum=0;
      for (i=1;i<=n;i++){
         err=(a*eig[i]-eig[i]*d[i]).abs();
       //  err*=eSupInv;
            // without this normalization very large values of the
            // components of a would (completely legally!) give large
            // numbers here and would suggest the presence of
            // irregularities which don't exist
         errCum+=err;
      }
      cpmmessage("test 1: err="&cpmwrite(errCum));
      cpmwait(tWait);
      R_Matrix eigt=~eig;
      R_Matrix unit=eig.net(1);
      R_Matrix e1=eig*eigt-unit;
      R_Matrix e2=eigt*eig-unit;
      err=e1.abs();
      errCum+=err;
      cpmmessage("test 2: err="&cpmwrite(err));
      cpmwait(tWait);
      err=e2.abs();
      errCum+=err;
      cpmmessage("test 3: err="&cpmwrite(err));
      cpmwait(tWait);
      cpmmessage("CpmArrays::eigen(...)  test done, errCum="&
         cpmwrite(errCum));
      cout<<"CpmArrays::eigen(...)  test done, errCum="<<errCum<<endl;
      cpmwait(tWait*2);
   }
   CPM_MZ
   return d;
}

#else

R_Vector CpmArrays::eigen(R_Matrix const& a, R_Matrix & eig,
   bool test, Z iterMax)
{
   const Z mL=2;
   static Word loc="CpmArrays::eigen(...)";
   CPM_MA
   Z i,n=a.dim1(),n1=a.dim2();
   if (n!=n1) cpmerror("CpmArrays::eigen(...): dimension mismatch");

   R_Vector d(n),e(n);

   eig=a; // unit matrix of same frame as a
   R eSup=eig.supAbs();
   if (eSup==R(0.)){
      eig = a.net(1);
      return d;
   }
   R eSupInv=R(1.)/eSup;
   eig*=eSupInv; // to avoid all trouble that the
   // tred2 and tqli-algorithms may have with extreme large or small
   // numbers. So far, there is no indication that there are problems
   // with large or small numbers.

   tred2(eig,d,e);
   tqli(d,e,eig,iterMax);
   V<Z> p=(-d).permutationForIncreasingOrder();
   d=d.permute(p);
   d*=eSup; // corrects for the scaling of the matrix

   eig=eig.permuteColumns(p);
   eig=~eig;
   for (i=1;i<=n;i++){
      Iv ise=eig[i].infSup();
      R vmin=ise.inf(),vmax=ise.sup();
      if (cpmabs(vmin)>cpmabs(vmax)) eig[i]*=R(-1.);
   }
   if (test){
      R tWait=1;
      cpmmessage("CpmArrays::eigen(...) test started");
      R aN=a.abs();
      cpmmessage("The norm of the matrix under consideration is  "&
         cpm(aN));
      R err, errCum=0;
      for (i=1;i<=n;i++){
         err=(a*eig[i]-eig[i]*d[i]).abs();
         err*=eSupInv;
            // without this normalization very large values of the
            // components of a would (completely legally!) give large
            // numbers here and would suggest the presence of
            // irregularities which don't exist
         errCum+=err;
      }
      cpmmessage("test 1: err="&cpmwrite(errCum));
      cpmwait(tWait);
      R_Matrix eigt=~eig;
      R_Matrix unit=eig.net(1);
      R_Matrix e1=eig*eigt-unit;
      R_Matrix e2=eigt*eig-unit;
      err=e1.abs();
      errCum+=err;
      cpmmessage("test 2: err="&cpmwrite(err));
      cpmwait(tWait);
      err=e2.abs();
      errCum+=err;
      cpmmessage("test 3: err="&cpmwrite(err));
      cpmwait(tWait);
      cpmmessage("CpmArrays::eigen(...)  test done, errCum="&
         cpmwrite(errCum));
      cout<<"CpmArrays::eigen(...)  test done, errCum="<<errCum<<endl;
      cpmwait(tWait*2);
   }
   CPM_MZ
   return d;
}
#endif

R CpmArrays::scalar_product(const R_Vector& x, const R_Vector& y,
   const R_Vector& d)
{
   R_Vector xm(x);
   R_Vector ym(y);
   R z=0.;
   R h=1;
   Z n=d.dim();
   for (Z i=1;i<=n;i++){
      z+=d[i]*(xm|ym);
      xm=xm.der(h);
      ym=ym.der(h);
   }
   return z;
}

void CpmArrays::adjust_scalar_product(const R_Vector& x, R a, R_Vector w,
   R_Vector& d)
{
   R h=1.,sum=0.;
   Z n=w.dim();
   R_Vector xm=x;
   for (Z i=1;i<=n;i++){
      sum+=(xm|xm)*w[i];
      xm=xm.der(h);
   }
   if (sum==R(0.)) cpmerror("adjust_scalar_product: sum=0");
   R b=a*a/sum;
   d=b*w;
}

R_Vector CpmArrays::projection(const R_Vector& x,
   const R_Matrix& f, const R_Vector& d, Z i)
{
   Z m=f.dim1();
   Z n=f.dim2();
   Z n1=x.dim();
   if (n!=n1) cpmerror("projection: dimension mismatch");
    R_Vector y(n1);
   for (Z ii=1;ii<=m;ii++){
      y=y+scalar_product(x,f[ii],d)*f[ii];
   }
    if (i==-1){
      return x-y;
   }
   else{
      return y;
   }
}

void CpmArrays::ortho_system(const R_Matrix& fs, const R_Vector& d,
   R_Matrix& os)
{
   Z i,j;
   R_Matrix f=fs;
   Z m=f.dim1();
   R a;
   for (i=1;i<=m;i++){
      a=scalar_product(f[i],f[i],d);
      if (a<=R(0.)) cpmerror(
         "ortho_system: vanishing norm of a input vector");
      a=R(1)/cpmsqrt(a);
      f[i]=a*f[i];
   }       // now the f[i] are unit vectors.
   R_Matrix g(m);
   for (i=1;i<=m;i++){
      for (j=1;j<=m;j++){
         g[i][j]=scalar_product(f[i],f[j],d);
      }
   }
   R_Matrix eig;
   R_Vector D=eigen(g,eig);
   for (i=1;i<=m;i++){
      if (D[i]<=0.) cpmerror("ortho_system: D[i]<=0");
   }
   f=eig*f;
   for (i=1;i<=m;i++){
      a=cpmsqrt(R(1)/D[i]);
      f[i]=a*f[i];
   }
   for (i=1;i<=m;i++){
      Iv ifr=f[i].infSup();
      R vmin=ifr.inf(),vmax=ifr.sup();
      if (cpmabs(vmin)>cpmabs(vmax)) f[i]=R(-1.)*(f[i]);
   }  // makes the dominant part positiv
   os=f;
}

R CpmArrays::testR_MatEig(const R_Matrix& a)
{
   R_Matrix sym=a+~a;
   sym*=R(0.5);
   R_Matrix eig=sym;
   R_Vector d=eigen(sym,eig);
   R_Matrix diag(d);
   R sum=0.;
   sum+=sym.dis(~eig*diag*eig);
   R_Matrix u=sym.net(1); // unit matrix
   sum+=((~eig)*eig).dis(u);
   sum+=(eig*(~eig)).dis(u);
   return sum;
}

////////////////////////// class SpcObjR /////////////////////////////////

SpcObjR::SpcObjR(const R_Matrix& a, bool enfSym):n_(a.dim1()),mat_(a),
   vec_(n_,n_), val_(n_)
{
   const Z mL=2;
   Word loc("SpcObjR(R_Matrix,bool)");
   CPM_MA
   if (enfSym) mat_ = (mat_ + ~mat_)*R(0.5);
   cpmassert(n_==mat_.dim2(),loc);
#if defined(CPM_USE_EIGEN) && !defined(CPM_QUAD)
   using namespace Eigen;
   Z i,j,k,l;
   Matrix<R,Dynamic,Dynamic> ea(n_,n_);
   for (i=1,k=0;i<=n_;++i,++k) for (j=1,l=0;j<=n_;++j,++l)
      ea(k,l) = mat_[i][j];
   SelfAdjointEigenSolver<Matrix<R,Dynamic,Dynamic> > ses(ea);
   Matrix<R,Dynamic,1> val=ses.eigenvalues();
   Matrix<R,Dynamic,Dynamic> vec=ses.eigenvectors();
   for (i=1,k=0;i<=n_;++i,++k) for (j=1,l=0;j<=n_;++j,++l)
      vec_[i][j]=vec(l,k);
   for (i=1,k=0;i<=n_;++i,++k) val_[i] = val(k);
#else
   val_=eigen(mat_,vec_);
#endif
   CPM_MZ
}

R_Matrix SpcObjR::func(F<R,R> const& f)const
{
   R_Matrix res(n_,n_); // dyadic representation, presumably faster
   for (Z i=1;i<=n_;++i) res+=R_Matrix(vec_[i],vec_[i])*f(val_[i]);
   return res;
}

namespace{
   R squareFunc(const R& x){return x*x;}
}

R SpcObjR::getError()const
{
   R sum1=0.;
   Z i,j;
   R nInv = 1./n_;
   for (i=1;i<=n_;++i) sum1+=(mat_*vec_[i] - vec_[i]*val_[i]).abs();
   sum1*=nInv;
   R sum2=0.;
   for (i=1;i<=n_;i++){
         for (j=1;j<=n_;j++){
            sum2+=cpmabs((vec_[i]|vec_[j]) - krn(i,j));
               //krn = Kronecker's delta
         }
   }
   sum2*=(nInv*nInv);
   R err3=(mat_*mat_-func(F<R,R>(squareFunc))).abs();
   err3*=(nInv*nInv);
   R res = sum1 + sum2 + err3;
   return res;
}
