//? cpmstate2.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


/*
   cpmstate2.cpp
   see cpmstate2.h

*/
#include <cpmstate.h>
#include <cpmstate2.h>
#include <cpmobservable.h>
#include <cpmimagingtools.h>
#include <cpmmultiimage.h>
#include <cpmfft.h>
#include <cpmgraph.h>
#include <cpmv.h>
#include <cpms.h>
#include <cpmrvector.h>
#include <cpmvectormore.h>


using CpmQuantumMechanics::QMBase;
using CpmQuantumMechanics::State;
using CpmQuantumMechanics::State2;
using CpmQuantumMechanics::State2Data;
using CpmQuantumMechanics::State2Trj;
using CpmQuantumMechanics::State2Trj2;
using CpmQuantumMechanics::Observable;


State2Data State2::d_;

////////////////////// class State2 /////////////////////////////////////

State2::State2(V<C> const& s1, V<C> const& s2):
   m_(s1.dim()),n_(s2.dim()),rep_(m_,V<C>(n_))
{
   Z i,j;
   for (i=1;i<=m_;++i){
      C si=s1[i];
      for (j=1;j<=n_;++j){
         C sij=si*s2[j];
         rep_.cui(i).cui(j)=sij;
      }
   }
   ini_();
}

State2::State2(State const& s1, State const& s2)
{
   V<C> sr1=s1.getRep();
   V<C> sr2=s2.getRep();
   *this=State2(sr1,sr2);
}

State2 State2::V12()const
{
   Z mL=3;
   Word loc("State2::V12()");
   CPM_MA
   Z i,j;
   State2 res(m_,n_);
   for (i=1;i<=m_;++i){
      R xi=x_[i];
      for (j=1;j<=n_;++j){
         R yj=y_[j];
         C repij=rep_.cui(i).cui(j);
         R rij=dis(xi,yj);
         res.rep_.cui(i).cui(j)=repij*v12(rij);
      }
   }
   CPM_MZ
   return res;
}

State2 State2::dot(R t)const
// new version with symmetric treatment of external potentials
// for 1 and 2 directions
{
   Z mL=3;
   Word loc("State2::dot()");
   CPM_MA
   Z i,j;
   R hDirac=QMBase::get_hDirac();
   R m1=d_.m1;
   cpmassert(m1!=0,loc);
   R m2=d_.m2;
   cpmassert(m2!=0,loc);
   cpmassert(m_>0,loc);
   cpmassert(n_>0,loc);
   cpmassert(d_.size1>R(0.),loc);
   cpmassert(d_.size2>R(0.),loc);
   R h1=R(d_.size1)/R(m_); // not m_-1 see cpmstate.h
      // comment to static Vl<R> positions(Z d);
   R h2=R(d_.size2)/R(n_);
   R f1=hDirac/h1;
   f1*=(-f1);
   R f2=hDirac/h2;
   f2*=(-f2);
   R fac1=f1/(m1*2);
   R fac2=f2/(m2*2);
   R facs=(fac1+fac2)*(-2);
   C fac=QMBase::get_ihDiracInv();
   State2 res(m_,n_);
   Outside box1=(Outside)box1_;
   Outside box2=(Outside)box2_;
   if (!d_.barrierForm){
      for (i=1;i<=m_;++i){
         R xi=x_[i];
         R v1=pot1(xi);
         for (j=1;j<=n_;++j){
            R yj=y_[j];
            C repij=rep_.cui(i).cui(j);
            C cij=repij*facs+
            (rep_.read(i-1,box1).read(j,box2)+rep_.read(i+1,box1).
               read(j,box2))*fac1+
            (rep_.cui(i).read(j-1,box2)+rep_.cui(i).read(j+1,box2))*fac2;
            R rij=dis(xi,yj);
            R v12=pot12(rij);
            R v2=pot2(yj);
            C vij=repij*(v1+v12+v2);
            cij+=vij;
            cij*=fac;
            res.rep_.cui(i).cui(j)=cij;
         }
      }
   }
   else{ // new form with dividing barrier
      V<R> p1(m_);
      for (i=1;i<=m_;++i) p1.cui(i)=pot1(x_.cui(i),t);

      V<R> p2(n_);
      for (j=1;j<=n_;++j) p2.cui(j)=pot2(y_.cui(j),t);

      for (i=1;i<=m_;++i){
         for (j=1;j<=n_;++j){
            C repij=rep_.cui(i).cui(j);
            C cij=repij*facs+
            (rep_.read(i-1,box1).read(j,box2)+rep_.read(i+1,box1).
               read(j,box2))*fac1+
            (rep_.cui(i).read(j-1,box2)+rep_.cui(i).read(j+1,box2))*fac2;
            C vij=repij*(p1.cui(i)+p2.cui(j));
            cij+=vij;
            cij*=fac;
            res.rep_.cui(i).cui(j)=cij;
         }
      }
   }
   CPM_MZ
   return res;
}

State2& State2::operator *=(C z)
{
   for (Z i=1;i<=m_;i++){
      rep_.cui(i).fc_(QMBase::prodFunc,z);
   }
   return *this;
}

State2& State2::operator *=(R r)
{
   for (Z i=1;i<=m_;i++){
      rep_.cui(i).fc_(QMBase::prodFuncReal,C(r));
   }
   return *this;
}

State2& State2::operator +=(const State2& psi)
{
   for (Z i=1;i<=m_;i++){
      rep_.cui(i).ff_(QMBase::sumFunc,psi.rep_.cui(i));
   }
   return *this;
}

State2& State2::operator -=(const State2& psi)
{
   for (Z i=1;i<=m_;i++){
      rep_.cui(i).ff_(QMBase::diffFunc,psi.rep_.cui(i));
   }
   return *this;
}

C State2::operator |(State2 const& psi)const
{
   C res;
   for (Z i=1;i<=m_;i++){
      res+=rep_.cui(i).fAcc2(QMBase::prodFuncConj1,
         QMBase::accFunc,psi.rep_.cui(i));
   }
   return res;
}

R State2::sup()const
{
   R res=-1;
   R val;
   Z i,j;
   for (i=1;i<=m_;i++){
      for (j=1;j<=n_;j++){
         val=rep_.cui(i).cui(j).abs();
         if (val>res) res=val;
      }
   }
   return res;
}

R State2::norm()
{
   return abs();
}

R State2::nor_(R r)
{
   R res=abs();
   if (res>0){
      R fac=R(1)/res;
      operator *=(fac);
   }
   return res;
}

void State2::con_()
 // complex conjugation, makes motion reversal
{
   Z i,j;
   for (i=1;i<=m_;i++){
      for (j=1;j<=n_;j++){
         rep_.cui(i).cui(j).con_();
      }
   }
}

void State2::concentrate_()
{
   Z i,j;
   for (i=1;i<=m_;i++){
      for (j=1;j<=n_;j++){
         R fac=rep_.cui(i).cui(j).absSqr();
         rep_.cui(i).cui(j)*=fac;
      }
   }
   nor_(1.);
}

void State2::backGrn_(R level)
      // changes the state in a new one for which to each component
      // the real level gets added. Finally normalizing
{
   C z(level);
   Z i,j;
   for (i=1;i<=m_;i++){
      for (j=1;j<=n_;j++){
         rep_.cui(i).cui(j)+=z;
      }
   }
   nor_(1.);
}

State2 State2::con(void)const
{
   cpmerror("State2::con(void)const not implemented");
   return State2();
}

void State2::mark(Graph& gr, Z ipolMethod, R gamma, R secFac)const
{
   Color::setGamma(gamma);
   bool aspRatFromState=false;
   bool periodic=true;
   gr.markVVC(rep_,false,Iv(),ipolMethod,aspRatFromState,periodic,secFac);
      // fast graphics function (made 2006-07-18)
}

State State2::red1(State const& psi)const
{
   Z mL=2;
   Word loc("State2::red1(State)");
   CPM_MA
   cpmassert(psi.dim()==m_,loc);
   State res(n_);
   for (Z j=1;j<=n_;++j){
      C cj;
      for (Z i=1;i<=m_;++i){
         cj+=psi.cui(i)*rep_.cui(i).cui(j);
      }
      res.cui(j)=cj;
   }
   CPM_MZ
   return res;
}

State State2::red2(State const& psi)const
{
   Z mL=2;
   Word loc("State2::red2(State)");
   CPM_MA
   cpmassert(psi.dim()==n_,loc);
   State res(m_);
   for (Z i=1;i<=m_;++i){
      C ci;
      for (Z j=1;j<=n_;++j){
         ci+=rep_.cui(i).cui(j)*psi.cui(j);
      }
      res.cui(i)=ci;
   }
   CPM_MZ
   return res;
}

void State2::smooth_()
// does not conserve the norm although the weights are
// optimum for doing so
{
   V< V<C> > rep(rep_);
   Z i1,i2,i1p,i1m,i2p,i2m;
   R fac=1./16;
   Outside box1=(Outside)box1_;
   Outside box2=(Outside)box2_;
   for (i1=1; i1<=m_; i1++){
      for (i2=1; i2<=n_; i2++){
         i1p=i1+1; i1m=i1-1;
         i2p=i2+1; i2m=i2-1;
         C s0=rep.cui(i1).cui(i2);
         C s1=rep.read(i1m,box1).read(i2,box2)+
            rep.read(i1p,box1).read(i2,box2);
         C s2=rep.cui(i1).read(i2m,box2)+
            rep.cui(i1).read(i2p,box2);
         C s3=rep.read(i1p,box1).read(i2m,box2)+
            rep.read(i1m,box1).read(i2p,box2);
         C s4=rep.read(i1p,box1).read(i2p,box2)+
            rep.read(i1m,box1).read(i2m,box2);
         C res=(s0*4.0+(s1+s2)*2.0+s3+s4)*fac;
         rep_.cui(i1).cui(i2)=res;
      }
   }
}

State2 State2::ran()const
{

   State2 res(*this);
   Z i1,i2;
   R r1(1.);
   C z0(r1,r1);
   for (i1=1; i1<=m_; ++i1){
      for (i2=1; i2<=n_; ++i2){
         res.put_(i1,i2,z0.ran());
      }
   }
   res.nor_();
   return res;
}

R State2::dtSup(Z p, R tWait, R gamma, R t)const
{
   Z mL=1;
   Word loc("State2::dtSup(Z,R,R,R)");
   CPM_MA
   cpmassert(p>0,loc);
   State2 psi=ran();
      // psi should not be special, especially not constant
      // in larger areas. Random unit vector is best.
   R_Vector mem(p);
   Graph gr;
   R pInv=R(1)/p;
   for (Z i=1;i<=p;++i){
      psi=psi.dot(t);
      R mi=psi.nor_(); // the value is the norm!
      cpmassert(mi>0.,loc);
      mem[i]=mi; // the value is the norm!
      cpmvalue("enr",mi,2);
      if (tWait>0){
         psi.mark(gr,0,gamma);
         gr.vis();
         cpmwait(tWait);
      }
      if (i==1){
         R n1=mem[i];
         cpmdebug(n1);
      }
      cpmprogress("x",R(i)*pInv,3);
   }
   Z ph=p/2;
   mem=mem.eliFirst(ph); // only the last half of the data
      // are used for determining the norm of the operator
      // psi |-> psi.dot()
   R_Vector stat=CpmArrays::statistics(mem);
   R sigma=stat[3];
   R maxVal=stat[10];
   cpmdebug(maxVal);
   R epsMax=maxVal+sigma;
   cpmdebug(epsMax);
   R res=R(2.)/epsMax; // notice the factor 2 is correct for finding
      // the stability limit of the direct midpoint method and the DALF
      // method. For the ALF method one has 1 instead.
   CPM_MZ
   return res;
}

//////////////// class State2Trj ///////////////////////////////////////

State2Trj::State2Trj(State2 const& psi, R dtRel, bool rel, R t0):
t_(t0), x_(psi),v_(x_.dot()),enfNor_(false),autoStep_(false), normTol_(-1),
stepMet_(0),revMet_(1)
{
   dtL_= rel ? dtLimit() : R(1.);
   dt_=  dtL_*dtRel;
      // if rel==false, a bit artificial case, the meaning of dtRel is
      // not that of a relative quantity, so the name is misleading
      // in this case.
}

void State2Trj::step_()
{
   Z mL=3;
   Word loc("State2Trj::step_()");
   CPM_MA
   R tau=dt_*0.5;
   R tauh=tau*0.5;
//   t_+=tau;
   if (stepMet_==-2){ // not a time stepping method,
      // only to visualize iterated application of dot
      t_+=dt_;
      v_=x_.dot();
      x_=v_;
   }
   else if (stepMet_==-1){ // Euler
      State2 acc=x_.dot(t_).dot(t_);
      x_+=(v_*dt_+acc*(dt_*tau));
      v_+=(acc*dt_);
      t_+=dt_;
   }
   else if (stepMet_==0){
      // lean implementation of the direct midpoint method
      t_+=tau;
      x_+=v_*tau;
      v_+=x_.dot(t_).dot(t_)*dt_;
      x_+=v_*tau;
      t_+=tau;
      phaseJitter_(jAmp1_,jAmp2_);
   }
   else if (stepMet_==-3){
      // Askar leap-frog (2014-10-05: can't reconstruct this reference
      // to Askar, isn't Askar's method normal leapfrog? Here we have
      // asynchronous leapfrog ALF, actually DALF)
      // Method of choice for time-dependent Hamiltonians
      t_+=tauh;
      x_+=v_*tauh;
      v_=x_.dot(t_)*2.-v_;
      x_+=v_*tau;
      t_+=tau;
      v_=x_.dot(t_)*2.-v_;
      x_+=v_*tauh;
      t_+=tauh;
      phaseJitter_(jAmp1_,jAmp2_);
   }
   else if (stepMet_==1){ // needs no v_ as data member
      t_+=tau;
      v_=x_.dot(t_);
      x_+=(v_*dt_);
      t_+=tau;
   }
   else if (stepMet_==2 || stepMet_==3){
      State2 xm=x_+v_*tau; // mid-point state, needs v_
      if (stepMet_==3){ // normal step, uses acceleration
         // old-fashioned implementation of the direct midpoint
         // method which is stepMet_==0 in modern form
         t_+=tau;
         State2 dv=xm.dot(t_).dot(t_)*dt_;
         State2 dx=v_*dt_+dv*tau;
         x_+=dx;
         v_+=dv;
         t_+=tau;
      }
      else if (stepMet_==2){ // uses velocity only
         t_+=tau;
         v_=xm.dot(t_);
         State2 dx=v_*dt_;
         x_+=dx;
         t_+=tau;
      }
   }
   else if (stepMet_==4){ // uses x_,v_,a_
      // method of choice for time independent Hamiltoians
           // lean implementation of the direct midpoint method, no jitter
      t_+=tau;
      x_+=v_*tau;
      v_+=x_.dot(t_).dot(t_)*dt_;
      x_+=v_*tau;
      t_+=tau;
   }
   else cpmerror("State2Trj::step_(): undefined step method");

   if (enfNor_){
      R xn=x_.nor_();
      if (xn==0.) cpmerror("zero norm in State2Trj::step_()");
      v_*=(R(1)/xn);
   }
   else if (normTol_>0){
      R xn=x_.abs();
      if (cpmabs(xn-1.)>normTol_){
         R xnInv=R(1)/xn;
         x_*=xnInv;
         v_*=xnInv;
      }
   }
   else if (autoStep_){ // not yet useful
/***********************
      R xn=x_.abs();
      if (xn>1.0001 || xn<0.95){ // xn>1 signals that Nyquist is not
      // satisfied
         // so this calls for action
         smooth_();
         dt_*=0.9;
         cpmmessage("dt_/dtL_="&cpm(dt_/dtL_));
      }
******************/
      // strategy is to hold dev in an acceptable range
      // Not yet reasonable.
      static R dU=1.e-4;
      static R dL=1.e-6;
      static R fDown=0.75;
      static R fUp=1.25;
      static R lev=1.;
      R xn=x_.abs();
      R dev=cpmabs(xn-lev);
      if (dev > dU) { dt_*=fDown; lev=xn;}
      else if (dev < dL) { dt_*=fUp; lev=xn;}
      else;
      cpmvalues2("dev,dt",dev,dt_,3);
   }
   else ;
   CPM_MZ
}

void State2Trj::run_(RunControl const& rc, Graph& gr)
{
   Z mL=1;
   Word loc("State2Trj::run_(RunControl,Graph)");
   CPM_MA
   gr.clr_();
   R ti=t_;
   R tf=rc.tTot;
   R ts=tf-ti; // time span
   Z nDoc=rc.nDoc;
   Z nImg=rc.nImg;
   Z nView=rc.nView;
   Z ipolMethod=rc.ipolMethod;
   R gamma=rc.gamma;
   cpmassert(gamma>0,loc);
   R tWait=rc.tWait;
   R dtDoc= nDoc<=1 ? 0. : ts/(nDoc-1);
   R dtImg= nImg<=1 ? 0. : ts/(nImg-1);
   R dtView= nView<=1 ? 0. : ts/(nView-1);
   PeriodicAlarm aTot(ts,ti,1.); // final 1 lets ti not fire
   PeriodicAlarm aDoc(dtDoc,ti);
   PeriodicAlarm aImg(dtImg,ti);
   PeriodicAlarm aView(dtView,ti);
   vector<R> v1,v2,v3,v4;
   R xx0=xx();
   C xv0=xv();
   cpmassert(xx0!=0.,loc);
   R xv0a=xv0.abs();
   cpmassert(xv0a!=0.,loc);
   R xx0AbsInv=R(1)/xx0;
   R xv0AbsInv=R(1)/xv0a;
   R nuRel;
   C epsRel;
// Notice that alarms with period <=0 never fire
// The initial state (prior to any step_() should be recorded
   if (aView(t_)){
      x_.mark(gr,ipolMethod, gamma);
      gr.vis();
   }
   if (aImg(t_)){
      x_.mark(gr,ipolMethod, gamma);
      gr.addText(Word("t=")&cpm(t_));
      gr.vis(true);
   }
   if (aDoc(t_)){
      nuRel=(xx()-xx0)*xx0AbsInv;
      epsRel=(xv()-xv0)*xv0AbsInv;
      v1.push_back(t_);
      v2.push_back(nuRel);
      v3.push_back(epsRel[1]);
      v4.push_back(epsRel[2]);
   }
   cpmassert(ts!=0.,loc);
   R tsInv=R(1.)/ts;
   while (!aTot(t_)){
      cpmprogress("",(t_-ti)*tsInv);
      step_();
         // the final state of the last step should be recorded
      if (aView(t_)){
         x_.mark(gr,ipolMethod, gamma);
         gr.vis();
      }
      if (aImg(t_)){
         x_.mark(gr,ipolMethod, gamma);
         gr.addText(Word("t=")&cpm(t_));
         gr.vis(true);
      }
      if (aDoc(t_)){
         nuRel=(xx()-xx0)*xx0AbsInv;
         epsRel=(xv()-xv0)*xv0AbsInv;
         v1.push_back(t_);
         v2.push_back(nuRel);
         v3.push_back(epsRel[1]);
         v4.push_back(epsRel[2]);
      }
   }
   R_Matrix doc(4);
   doc[1]=R_Vector(v1);
   doc[2]=R_Vector(v2);
   doc[3]=R_Vector(v3);
   doc[4]=R_Vector(v4);
   gr.clr_();
   V<Color> vc{BLUE,GREEN,RED};
   Word leg("blue: nuRel, green: Re(epsRel), red: Im(epsRel)");
   gr.mark1(doc,vc,leg,0.25);
   gr.vis(true);
   cpmwait(tWait,2);
   CPM_MZ
}

R State2Trj::dtLimit(bool light)const
{
   if (light){
      R ds1=x_.ds1(),ds2=x_.ds2(),m1=x_.m1(),m2=x_.m2();
      R fac=R(2)/QMBase::get_hPlanck();
      R r1=ds1*ds1*m1;
      R r2=ds2*ds2*m2;
      return fac*CpmRootX::inf<R>(r1,r2);
   }
   else{
      Z iter=100;
      R tWait=0., gamma=1.;
      return x_.dtSup(iter,tWait,gamma,t_);
   }
}

void State2Trj::phaseJitter_(R relAmp1,R relAmp2)
      // multiplies each component by exp(i*phi) with -phi0<=phi<=phi0
      // and phi0=2*pi*relAmpl;
{
   if (relAmp1<=0. && relAmp2<=0.) return;
   Z m=x_.dim1();
   Z n=x_.dim2();
   V<C> phi1(m,C(1));
   V<C> phi2(n,C(1));
   Z i,j;
   if (relAmp1>0){
      R a1=cpmpi*relAmp1;
      Iv iv1(-a1,a1);
      for (i=1;i<=m;++i) phi1.cui(i)=C::expi(iv1.ran());
   }
   if (relAmp2>0){
      R a2=cpmpi*relAmp2;
      Iv iv2(-a2,a2);
      for (j=1;j<=n;++j) phi2.cui(j)=C::expi(iv2.ran());
   }
   for (i=1;i<=m;i++){
      C ci=phi1[i];
      for (j=1;j<=n;j++){
         C cij=ci*phi2[j];
         x_.put_(i,j,x_(i,j)*cij);
         v_.put_(i,j,v_(i,j)*cij);
      }
   }
}

//////////////// class State2Trj2 ///////////////////////////////////////
/*

State2Trj2::State2Trj2(State const& x1, State const& x2):
State2Trj(State2(x1,x2)),x1_(x1,TrjCon()),x2_(x2,TrjCon())
{}


void State2Trj2::step_()
{
   x_=State2(x1_,x2_);
   v_=x_.dot();
   State2Trj::step_();
   State x1m=x1_.step_();
   State x2m=x2_.step_();
   State x1=x_.red2(x2m);
   State x2=x_.red1(x1m);
   x1.nor_();
   x2.nor_();
   x1_=x1;
   x2_=x2;
}


void State2Trj2::step_()
{
   X2<State,State> dpsi=dot();
   x1_+=dpsi.c1()*dt_;
   x2_+=dpsi.c2()*dt_;
   x1_.nor_();
   x2_.nor_();
   x_=State2(x1_,x2_);
}

void State2Trj2::mark(Graph& f, Z ipolMeth, R gamma, R secFac)const
{
   Z mL=2;
   Word loc("State2Trj2::mark(Graph,Z,R,R)");
   CPM_MA
   Frames frs(f,1,2);
   Graph g1(frs(1));
   x1_.x().showAsCurves(g1);
   Graph g2(frs(2));
   x2_.showAsCurves(g2);
   cpmwait(0.5);
   CPM_MZ
}


X2<State, State> State2Trj2::dot()const
{
   Z mL=2;
   Word loc("State2Trj2::dot()");
   CPM_MA
   Z i,j;
   R hDirac=QMBase::get_hDirac();
   State2Data d_=x_.get_d();
   R m1=d_.m1;
   R m2=d_.m2;
   Z d1=x_.d1();
   Z d2=x_.d2();
   R h1=d_.size1/d1;
   R h2=d_.size2/d2;
   R f1=hDirac/h1;
   f1*=(-f1);
   R f2=hDirac/h2;
   f2*=(-f2);
   R fac1=f1/(2*m1);
   R fac2=f2/(2*m2);
   C fac=QMBase::get_ihDiracInv();

   State y1(d1);
   State y2(d2);
   Outside box1=DEFAULT;
   Outside box2=DEFAULT;

   for (i=0;i<d1;++i){
      R xi=x_.x(i);
      R v1_i=x_.pot1(xi);
      C x1_i=x1_.cui(i);
      C y1_i=fac1*(x1_.read(i-1,box1)-2*x1_i+x1_.read(i+1,box1));
      y1_i+=x1_i*v1_i;
      C sum_i;
      for (j=0;j<d2;++j){
         R yj=x_.y(j);
         R rij=x_.dis(xi,yj);
         R vij=x_.pot12(rij);
         C x2_j=x2_.cui(j);
         sum_i+=x2_j.absSqr()*vij;
      }
      sum_i*=x1_i;
      y1_i+=sum_i;
      y1.cui(i)=fac*y1_i;
   }

   for (j=0;j<d2;++j){
      R yj=x_.y(j);
      R v2_j=x_.pot2(yj);
      C x2_j=x2_.cui(j);
      C y2_j=fac2*(x2_.read(j-1,box2)-2*x2_j+x2_.read(j+1,box2));
      y2_j+=x2_j*v2_j;
      C sum_j;
      for (i=0;i<d1;++i){
         R xi=x_.x(i);
         R rij=x_.dis(xi,yj);
         R vij=x_.pot12(rij);
         C x1_i=x1_.cui(i);
         sum_j+=x1_i.absSqr()*vij;
      }
      sum_j*=x2_j;
      y2_j+=sum_j;
      y2.cui(j)=fac*y2_j;
   }
   CPM_MZ
   return X2<State,State>(y1,y2);
}
*/
