//? cpmvc.cpp
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


/*
   cpmvc.cpp
*/

#include <cpmvc.h>
#include <cpmvr.h>
using CpmArrays::VC;
using CpmArrays::Vr;

////////////////////////////// class VC /////////////////////////////
namespace{
   void accFunc(C& z, C const& zAdd){ z+=zAdd;}
   C conjFunc(C const& z){ return z.con();}
   C negFunc(C const& z1){  return -z1;}
   C sumFunc(C const& z1, C const& z2){ return z1+z2;}
   C diffFunc(C const& z1, C const& z2){ return z1-z2;}
   C prodFunc(C const& z1, C const& z2){ return z1*z2;}
   C prodFuncConj1(C const& z1, C const& z2){return z1.con()*z2; }
   C prodFuncConj2(C const& z1, C const& z2){ return z1*z2.con();}
   C prodFuncReal(C const& z1, C const& z2){ return z1*z2.real();}
}

C VC::operator |(VC const& x2)const
{
   return rep_.fAcc2(prodFuncConj1,accFunc,x2.rep_);
}

VC VC::con(void)const
{
   return VC(rep_.fa(conjFunc));
}

VC VC::operator - (void)const
{
   return VC(rep_.fa(negFunc));
}

namespace{
   C fpFunc(C const& z, R const& p)
   {
      return p==2 ? C(z.absSqr()) : C(cpmpow(z.abs(),p));
   }
}

VC& VC::makeReal_(R p)
{
#ifdef CPM_Fn
   F<C,C> fp=F1<C,R,C>(p)(fpFunc);
#else
   F<C,C> fp(bind(fpFunc,_1,p));
#endif
   rep_.fa_(fp,rep_.dom());
   return *this;
}

VC& VC::operator +=(VC const& x)
{
   rep_.ffl_(sumFunc,x.rep_);
   return *this;
}

VC& VC::operator -=(VC const& x)
{
   rep_.ffl_(diffFunc,x.rep_);
   return *this;
}

VC& VC::operator *=(C z)
{
   rep_.fc_(prodFunc,z);
   return *this;
}

VC& VC::operator *=(R r)
{
   rep_.fc_(prodFuncReal,C(r));
   return *this;
}

VC VC::operator +(VC const& x)const
{
  VC res(*this);
  return res+=x;
}

VC VC::operator -(VC const& x)const
{
   VC res(*this);
   return res-=x;
}

VC VC::operator *(C x2)const
{
  VC res(*this);
  return res*=x2;
}

VC VC::operator *(R x2)const
{
  VC res(*this);
  return res*=x2;
}

VC VC::ran(Z j)const
// combining ran from Vr<C> with normalization
// exemplifies the interaction of Vl<C> and Vr<C>.
// Actually also the arithmetics could  be taken
// from Vr<C> if a single source of algorithms
// is desired. If so, it would be more natural to
// use Vr<C> directly instead of VC.
{
   Vr<C> temp=Vr<C>(rep_);
   temp=temp.ran(j);
   V<C> r=temp;
   VC res=VC(r);
   res.nor_();
   return res;
}
