//? cpmvr.h
//? Status of work 2022-10-05.
//? //? Copyright (c) 2022 Ulrich Mutze
//? contact: see contact-info at www.ulrichmutze.de
//? 
//? This program is free software: you can redistribute it and/or
//? modify it under the terms of the GNU General Public License as
//? published by the Free Software Foundation, either version 3 of
//? the License, or (at your option) any later version.
//? 
//? This program is distributed in the hope that it will be useful,
//? but WITHOUT ANY WARRANTY; without even the implied warranty of
//? MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//? GNU General Public License <http://www.gnu.org/licenses/> for 
//? more details.


//?  ... 


#ifndef CPM_VR_H_
#define CPM_VR_H_
/*

Purpose: Defining a vector template providing the full
functionality of the rich interface

*/
#include <cpmva.h>

//////////////////////////// class Vr<> //////////////////////////////////

namespace CpmArrays{

   using CpmRoot::R;
   using CpmRoot::Root;

template <class T>
   // We assume that T provides (explicitely or implicitely)
   // copy constructor, and assignement
   // or that T is a built-in type.

class Vr: public Va<T>{ // version of Va with a rich interface

public:
   typedef Vr<T> Type;
   typedef Va<T> Base;
   typedef V<T> Source;
   typedef T ScalarType; // was missing till 2001-10-12
      // detected by the gnu compiler
   //CPM_RVR(Vr)

   explicit Vr(Z n=0):Va<T>(n){}
      // note that Vr(0) is defined and is different from Vr()

   explicit Vr(IvZ ivz):Va<T>(ivz){}
      // note that Vr(0) is defined and is different from Vr()

   Vr(IvZ ivz, T const& t):Va<T>(ivz,t){}

   // constructors from explicit lists
   explicit  Vr(std::initializer_list<T> il ):Va<T>(il){}
   // requires C++11
      // constructors from explicit lists such as
      // Vr<Z> v{1,2,4,8};

   Vr(Z n, T const& t):Va<T>(n,t){}
      // initializes all n components with t

   Vr( V<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor

   Vr( Vo<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor

   Vr(Va<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor
   Vr(Vr<T> const& h):Va<T>(h){}
      // 'down-cast'-constructor
/*
#if defined(CPM_USE_MOVE)
   Vr(Vr<T> && h)
   {
      Source::iv_=h.iv_;
      Source::sz_=h.sz_;
      Source::p_=h.p_;
      cout<<"move constructor Vr"<<endl;
      h.iv_=IvZ();
      h.sz_=0;
      h.p_=nullptr;
   }
#endif
*/

   virtual V<T>* clone(void)const{ return new Vr<T>(*this);}

   CPM_TEST
   CPM_DESCRIPTORS
      // dis from Va
};

// member functions

template <class T>
Word Vr<T>::nameOf(void)const
{
   Word w1="Vr<";
   Word w2=Root<T>(T()).nameOf();
   return w1&w2&">";
}

template <class T>
Word Vr<T>::toWord(void)const
{
   if (Base::isVoid()) return "void list";
   Word res="( ";
   Z i;
   for (i=Base::b();i<Base::e();i++){
      Word wi="("&Root<Z>(i).toWord()&","&Root<T>((*this)[i]).toWord()&"),\
       ";
      res&=wi;
   }
   i=Base::e();
   Word we="("&Root<Z>(i).toWord()&","&Root<T>((*this)[i]).toWord()&")";
   res&=we;
   res&=" )";
   return res;
}

template <class T>
Z Vr<T>::hash(void)const
{
   Z res=0;
   for (Z i=Base::b(); i<=Base::e(); i++){
      res+=Root<T>((*this)[i]).hash();
   }
   return res;
}

template <class T>
Vr<T> Vr<T>::test(Z cpl)const
{
   Z it=0;
   it=Root<Z>(it).test(cpl);
   T xt;
   xt=Root<T>(xt).test(cpl);
   return Vr<T>(it,xt);
}

template <class T>
Vr<T> Vr<T>::ran(Z j)const
   // all randomized versions of *this thus have the same dimension as
   // *this
{
   Vr<T> res(Base::dom());
   T tempX;
   if (j==0){
      for (Z i=Base::b(); i<=Base::e(); i++){
         tempX=(*this)[i];
         res[i]=Root<T>(tempX).ran(j);
      }
   }
   else{
      Z j0=1000; // this is done to avoid an obvious correlation
         // among the coordinates of x.ran(i)
         // and x.ran(i+1)
      Z jIncr=Root<Z>(j0).ran(j);
      Z jc=j;
      for (Z i=Base::b(); i<=Base::e(); i++){
         tempX=(*this)[i];
         jc+=jIncr;
         res[i]=Root<T>(tempX).ran(jc);
      }
   }
   return res;
}

template <typename T, typename S>
class Vrs: public Vr<T>{  // r: rich, s: scalar
      public:
         Vrs(){} 
         Vrs(Vr<T> const& vr):Vr<T>(vr){} 
         Vrs(Z n):Vr<T>(n){} 
         Vrs(Z n, T const& t):Vr<T>(n,t){} 
         Vrs& operator*=(S const& s);
         Vrs operator*(S const& s)const override;
     //    Vrs& operator*(S const& s);          
   };

template <typename T, typename S>
Vrs<T,S> Vrs<T,S>::operator*(S const& s )const
{
   Vrs<T,S> res(*this);
   Z n=res.dim();
   for (Z i=1;i<=n;++i){ 
      res.cui(i)*=s;
   }
   return Vrs(res);
}

//template <typename T, typename S>
//Vrs<T,S>& Vrs<T,S>::operator*(S const& s )
//{
//   Vrs<T,S> res(*this);
//   Z n=res.dim();
//   for (Z i=1;i<=n;++i) res.cui(i)*=s;
//   return Vrs(res);
//} 

template <typename T, typename S>
Vrs<T,S>& Vrs<T,S>::operator*=(S const& s )
{
   Z n=Vr<T>::dim();
   for (Z i=1;i<=n;++i) Vr<T>::cui(i)*=s;
   return *this;
}   

// We do not implement the iterated templates since they are now
// endowed with sufficient functionality already as Va, VVa, VVVa,
// and VVVVa

} // namespace

#endif
